/* add.c - ldif backend add function */
/* $OpenLDAP: pkg/ldap/servers/slapd/back-ldif/add.c,v 1.6.2.3 2000/06/13 17:57:37 kurt Exp $ */
/*
 * Copyright 1998-2000 The OpenLDAP Foundation, All Rights Reserved.
 * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
 */

#include "portable.h"

#include <stdio.h>

#include <ac/string.h>
#include <ac/socket.h>

#include "slap.h"
#include "slapd_ldif.h"

int
ldif_back_add(
    Backend	*be,
    Connection	*conn,
    Operation	*op,
    Entry	*e
)
{
	FILE			*rfp, *wfp;
	int			len;
	char			*binddn = NULL;
	char			*arguments[2];
        LdifDT *ldt = (LdifDT *) be->be_private;

	if ( ldt->ldifcmdadd == NULL ) {
		send_ldap_result( conn, op, LDAP_UNWILLING_TO_PERFORM, NULL,
		    "add not implemented", NULL, NULL );
		return( -1 );
	}
	arguments[0] = ldt->ldifcmdadd;
	arguments[1] = NULL;

	if ( (op->o_private = (void *) ldif_forkandexec( arguments, &rfp, &wfp )) == (void *) -1 ) {
		send_ldap_result( conn, op, LDAP_OPERATIONS_ERROR, NULL,
		    "could not fork/exec", NULL, NULL );
		return( -1 );
	}

	/* write out the request to the add process */
	fprintf( wfp, "ADD\n" );
	fprintf( wfp, "msgid: %ld\n", (long) op->o_msgid );
	print_ldif_suffixes( wfp, be );
	fprintf( wfp, "connid: %lu\n", conn->c_connid );
	ldap_pvt_thread_mutex_lock( &entry2str_mutex );
	fprintf( wfp, "%s", entry2str( e, &len ) );
	ldap_pvt_thread_mutex_unlock( &entry2str_mutex );

	if (conn->c_dn != NULL) {
		binddn = strchr(conn->c_dn, '=');
		if (binddn && strlen(binddn) > 1)
			fprintf( wfp, "bind_dn: %s\n", binddn+1);
		else
			fprintf( wfp, "bind_dn: NULL\n");
	}
	else
		fprintf( wfp, "bind_dn: NULL\n");

	fclose( wfp );

	/* read in the result and send it along */
	read_and_send_ldif_results( be, conn, op, rfp, NULL, 0 );

	fclose( rfp );
	return( 0 );
}
