/* config.c - ldif backend configuration file routine */
/*
$Header: /home/globdev/CVS/globus-packages/mds/openldap/ldapmodules/config.c,v 1.4.4.1 2003/04/17 19:33:57 jslee Exp $
$Log: config.c,v $
Revision 1.4.4.1  2003/04/17 19:33:57  jslee
fixing typo get_local_port

Revision 1.4  2003/03/27 03:08:45  jslee
allowing multiple grises in one slapd

Revision 1.3  2003/03/22 02:49:56  jslee
adding allowstaledata yes/no

Revision 1.2  2002/08/14 22:40:30  benc
Merge MDS2.2 into trunk.

Revision 1.1.1.1.28.1  2002/08/07 03:56:05  benc
Merged (hopefully) all of the mds_2_2_branch work onto this new branch.

Compiles and runs on Linux.

Revision 1.1.1.1.10.1  2002/05/20 21:44:35  jslee
adding syslog messages to the ldif backend

Revision 1.1.1.1  2002/01/10 02:10:09  jslee
importing globus-beta-branch to the main trunk

Revision 1.1.6.2  2001/10/13 03:45:33  jslee
adding GlobusStub object as default

Revision 1.1.6.1  2001/09/28 07:23:16  jslee
resolving naming conflicts with giis backend routines

Revision 1.1  2001/08/06 22:15:37  bletzing
New Package.  Doesn't Build.

Revision 1.1.1.1  2001/07/06 21:11:15  mei
1) splitting out back-ldif from the openldap 2.0.8-patched into a standalone
   directory that can generate 'dynamic modules' that can be used by the
   openldap's slapd (checked against 2.0.11 already)

Revision 1.3  2001/05/24 01:05:06  june
gridmap checking & timelimit debugging

Revision 1.2  2001/04/04 21:42:07  june
debugging add operation

Revision 1.1  2001/03/24 00:08:36  june
adding ldif backend

Revision 1.1  2001/03/22 03:01:25  june
add ldif backend

Revision 1.3  2000/09/13 02:11:50  june
resolve changes between 2.0 & 2.0.1

Revision 1.2  2000/09/12 22:41:39  june
Add ldif backend
*/



#include "portable.h"

#include <stdio.h>

#include <ac/socket.h>
#include <ac/string.h>
#include <ac/time.h>

#include "slap.h"
#include "external.h"
#include "slapd_ldif.h"

int
ldif_back_db_config(
    BackendDB	*be,
    const char	*fname,
    int		lineno,
    int		argc,
    char	**argv
)
{
	char *stub_obj;

	LdifDT *ldt = (LdifDT *) be->be_private;

	if (ldt == NULL) {
		ldt = (LdifDT *) calloc(1, sizeof(LdifDT));
		if (ldt == NULL) {
			fprintf(stderr, "ldif data memory allocation error\n");
			Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
			return -1;
		}
		if( init_ldifdt(ldt) == -1 )
			return -1;
		be->be_private = ldt;
	}

	if (ldt->ldifgiven_sfx != NULL) free(ldt->ldifgiven_sfx);
	if ( (ldt->ldifgiven_sfx = (char *) strdup(be->be_suffix[0])) == NULL) {
		fprintf(stderr, "memory allocation error\n");
		Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
		return( -1 );
	}


	if (!ldt->stub_entry) {
		stub_obj = (char *) calloc(strlen("dn: ") + strlen("objectclass: ") + strlen(be->be_suffix[0]) + strlen("GlobusStub") + 3, sizeof(char));
		if(!stub_obj) {
			fprintf(stderr, "memory allocation error\n");
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			return -1;
		}
		sprintf(stub_obj, "dn: %s\nobjectclass: GlobusStub\n", be->be_suffix[0]);
		ldt->stub_entry = str2entry( stub_obj );
		if(!ldt->stub_entry) {
			fprintf(stderr, "str2entry error\n");
			Debug (LDAP_DEBUG_ANY, "str2entry error\n", 0, 0, 0);
			return -1;
		}
	}

	/* ldif resource script */
	if ( strcasecmp( argv[0], "search" ) == 0 ) {

		if ( argc < 2 ) {
			fprintf( stderr,
		"%s: line %d: missing filename in \"search <filename>\" line\n",
			    fname, lineno );
			Debug( LDAP_DEBUG_ANY, "%s: line %d: missing filename in \"search <filename>\" line\n", fname, lineno, 0 );
			return( 1 );
		}

	} else if ( strcasecmp( argv[0], "add" ) == 0 ) {
		if ( argc < 2 ) {
			fprintf( stderr,
		"%s: line %d: missing command name in \"add <cmdname>\" line\n",
			    fname, lineno );
			Debug( LDAP_DEBUG_ANY, "%s: line %d: missing command name in \"add <cmdname>\" line\n", fname, lineno, 0 );
			return( 1 );
		}
		if (ldt->ldifcmdadd) free(ldt->ldifcmdadd);
		ldt->ldifcmdadd = ch_strdup( argv[1] );
		if (!ldt->ldifcmdadd) {
			fprintf(stderr, "memory allocation error\n");
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			return 1;
		}

	} else if ( strcasecmp( argv[0], "anonymousbind" ) == 0 ) {
		if ( argc < 2 ) {
			fprintf( stderr,
		"%s: line %d: missing command name in \"anonymousbinding <cmdname>\" line\n",
			    fname, lineno );
			Debug( LDAP_DEBUG_ANY, "%s: line %d: missing command name in \"anonymousbinding <cmdname>\" line\n", fname, lineno, 0 );
			return( 1 );
		}
		if (argv[1] && !strcasecmp(argv[1], "no"))
			ldt->ldifallow_anonym = 0;
		else
			ldt->ldifallow_anonym = 1;

	} else if ( strcasecmp( argv[0], "allowstaledata" ) == 0 ) {
		if ( argc < 2 ) {
			fprintf( stderr,
		"%s: line %d: missing command name in \"allowstaledata <cmdname>\" line\n",
			    fname, lineno );
			Debug( LDAP_DEBUG_ANY, "%s: line %d: missing command name in \"allowstaledata <cmdname>\" line\n", fname, lineno, 0 );
			return( 1 );
		}
		if (argv[1] && !strcasecmp(argv[1], "no"))
			ldt->allow_stale_data = 0;
		else
			ldt->allow_stale_data = 1;

	} else if ( strcasecmp( argv[0], "conf" ) == 0 ) {
		if ( argc < 2 ) {
			fprintf( stderr,
		"%s: line %d: missing filename in \"conf <filename>\" line\n",
			    fname, lineno );
			Debug( LDAP_DEBUG_ANY, "%s: line %d: missing filename in \"conf <filename>\" line\n", fname, lineno, 0 );
			return( 1 );
		}
		if (ldt->ldifconfigfile) free(ldt->ldifconfigfile);
		ldt->ldifconfigfile = ch_strdup( argv[1] );
		if (!ldt->ldifconfigfile) {
			fprintf(stderr, "memory allocation error\n");
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			return 1;
		}
		if (ldif_readconf(ldt) == -1) {
			fprintf(stderr, "config file (%s) read error\n", ldt->ldifconfigfile);
			Debug(LDAP_DEBUG_ANY, "config file (%s) read error\n", ldt->ldifconfigfile, 0, 0 );
			return 1;
		}

	/* anything else */
	} else {
		fprintf( stderr,
"%s: line %d: unknown directive \"%s\" in ldif database definition (ignored)\n",
		    fname, lineno, argv[0] );
		Debug( LDAP_DEBUG_ANY, "%s: line %d: unknown directive \"%s\" in ldif database definition (ignored)\n", fname, lineno, argv[0] );
	}

	return( 0 );
}

int init_ldifdt(LdifDT *ldt)
{

	ldt->glist = (ActiveObj ***) calloc(1, sizeof(ActiveObj **));
	if(!ldt->glist) {
		fprintf(stderr, "memory allocation error\n");
		Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
		return -1;
	}
	*ldt->glist = NULL;
	ldt->total_inx = 0;
	ldap_pvt_thread_mutex_init( &(ldt->glist_mutex) );

	ldt->ldifconfigfile = NULL;
	ldt->ldifcmdadd = NULL;
	ldt->ldifgiven_sfx = NULL;
	ldt->ldifallow_anonym = 1;
	ldt->allow_stale_data = 1;
	ldt->local_port = ldif_get_local_port();
	ldt->stub_entry = NULL;

	return 0;
}

