/* init.c - initialize ldif backend */
/*
$Header: /home/globdev/CVS/globus-packages/mds/openldap/ldapmodules/init.c,v 1.7.4.1 2003/07/03 01:03:29 olomu Exp $
$Log: init.c,v $
Revision 1.7.4.1  2003/07/03 01:03:29  olomu
Performance modifications bugfixes committed to 2_4_branch

Revision 1.8  2003/05/07 20:08:04  olomu
GRIS Update for performance.  Using linked structure of Openldap Entries for caching GRIS data

Revision 1.7  2003/03/27 03:08:45  jslee
allowing multiple grises in one slapd

Revision 1.6  2003/03/18 18:14:20  meder
Globus builds on darwin now

Revision 1.5  2003/02/06 01:55:33  jslee
fixing typo - ldif_getstrent

Revision 1.4  2003/02/05 09:10:40  jslee
fixing threads runout - bugzilla #644

Revision 1.3  2002/10/16 01:18:03  jslee
fixing file descriptor leaks

Revision 1.2  2002/08/14 22:40:30  benc
Merge MDS2.2 into trunk.

Revision 1.1.1.1.28.1  2002/08/07 03:56:05  benc
Merged (hopefully) all of the mds_2_2_branch work onto this new branch.

Compiles and runs on Linux.

Revision 1.1.1.1.10.1  2002/05/20 21:44:35  jslee
adding syslog messages to the ldif backend

Revision 1.1.1.1  2002/01/10 02:10:09  jslee
importing globus-beta-branch to the main trunk

Revision 1.1.6.5  2001/10/30 01:26:26  jslee
fixing concurrency problems

Revision 1.1.6.4  2001/10/11 21:24:08  jslee
fixing timelimit, sizelimit routines

Revision 1.1.6.3  2001/09/28 07:23:16  jslee
resolving naming conflicts with giis backend routines

Revision 1.1.6.2  2001/09/26 21:18:38  smartin
committing merge from main repository trunk to beta branch

Revision 1.2  2001/09/24 15:13:10  smartin
committing merge from globus-beta-branch

Revision 1.1.6.1  2001/09/21 19:58:35  meisu
1) rename ldif.h to slapd_ldif.h
2) alter the Makefile.am to match with giis-ldapmodule's and the
   flavor is in the install path

Revision 1.1  2001/08/06 22:15:38  bletzing
New Package.  Doesn't Build.

Revision 1.1.1.1  2001/07/06 21:11:15  mei
1) splitting out back-ldif from the openldap 2.0.8-patched into a standalone
   directory that can generate 'dynamic modules' that can be used by the
   openldap's slapd (checked against 2.0.11 already)

Revision 1.4  2001/05/24 01:05:06  june
gridmap checking & timelimit debugging

Revision 1.3  2001/04/15 00:56:05  june
apply env reading routine to dn

Revision 1.2  2001/04/11 23:05:27  june
add env variable reading routines

Revision 1.1  2001/03/24 00:08:37  june
adding ldif backend

Revision 1.1  2001/03/22 03:01:25  june
add ldif backend

Revision 1.3  2000/09/13 02:11:50  june
resolve changes between 2.0 & 2.0.1

Revision 1.2  2000/09/12 22:41:39  june
Add ldif backend
*/


#include "portable.h"

#include <stdio.h>

#include <ac/socket.h>

#include <ac/signal.h>

#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <ctype.h>
#include <sys/socket.h>
#include <string.h>
#include <errno.h>

#include <strings.h>
#include <stdlib.h>

#define ENV_VAR_SIZE 1
#define SLAPD_LDIF_DYNAMIC 1
#define MAX_HOSTNAME 1024

char *ldif_apply_env_vars(char *);
char *ldif_str_realloc(char *, int, int *);

#include "slap.h"
#include "slapd_ldif.h"

#ifdef SLAPD_LDIF_DYNAMIC

int libback_ldif_LTX_init_module(int argc, char *argv[]) {
    BackendInfo bi;

    memset( &bi, '\0', sizeof(bi) );
    bi.bi_type = "ldif";
    bi.bi_init = ldif_back_initialize;

    backend_add(&bi);
    return 0;
}


#endif /* SLAPD_LDIF_DYNAMIC */

int
ldif_back_initialize(
    BackendInfo	*bi
)
{
	bi->bi_open = 0;
	bi->bi_config = 0;
	bi->bi_close = 0;
	bi->bi_destroy = 0;

	bi->bi_db_init = 0;
	bi->bi_db_config = ldif_back_db_config;
	bi->bi_db_open = 0;
	bi->bi_db_close = 0;
	bi->bi_db_destroy = 0;

	bi->bi_op_bind = ldif_back_bind;
	bi->bi_op_unbind = 0;
	bi->bi_op_search = ldif_back_search;
	bi->bi_op_compare = 0;
	bi->bi_op_modify = ldif_back_modify;
	bi->bi_op_modrdn = ldif_back_modrdn;
	bi->bi_op_add = ldif_back_add;
	bi->bi_op_delete = ldif_back_delete;
	bi->bi_op_abandon = ldif_back_abandon;

	bi->bi_extended = 0;

	bi->bi_acl_group = 0;
	bi->bi_acl_attribute = 0;
	bi->bi_chk_referrals = 0;

	bi->bi_connection_init = 0;
	bi->bi_connection_destroy = 0;

	return 0;
}


int 
ldif_readconf(LdifDT *ldt)
{

	ActiveObj	*current = NULL, ***glist = ldt->glist;
	int		no, data_size = LDIF_DATA_SIZE, inx;
	int		activeobj, activesch, objclass_no, objclass_inx, obj_no;
	int		toggle, type, cachetime, timelimit, sizelimit;
	char		*typestr, *cachetimestr, *timelimitstr, *sizelimitstr;
	FILE		*fptr;
	char		*tmpobj, *obj, *path, *base, *args, *dn, *tmp;
	char		**objclass, *tpath, *targs, *tdn;
	Entry           *e;
	static char	*objtypes[] = {"EXEC", "DYNAMIC", "STATIC", "INVALID"};

	if ( ldt->ldifconfigfile == NULL ) {
		fprintf( stderr, "filename not found\n" );
		Debug (LDAP_DEBUG_ANY, "filename not found\n", 0, 0, 0);
		return( -1 );
        }

	if ( (fptr = fopen( ldt->ldifconfigfile, "r" )) == NULL ) {
		fprintf (stderr, "Can't open %s\n", ldt->ldifconfigfile);
		Debug (LDAP_DEBUG_ANY, "Can't open %s\n", ldt->ldifconfigfile, 0, 0 );
		return -1;
	}

	ldap_pvt_thread_mutex_lock(&ldt->glist_mutex);
	if (*glist) {
		ldif_clear_list(*glist, ldt->total_inx);
		ldt->total_inx = 0;
	}
	else {
		*glist = (ActiveObj **) calloc(data_size, sizeof(ActiveObj *));
		if (!*glist) {
			fprintf( stderr, "memory allocation failed\n" );
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			fclose(fptr);
			return( -1 );
		}
		ldt->total_inx = 0;
	}
	ldap_pvt_thread_mutex_unlock(&ldt->glist_mutex);

	for ( obj = ldif_getstrent(fptr); obj != NULL; obj = ldif_getstrent(fptr) ) {

		tdn = ldif_get_attribute(obj, "dn:", NULL);
		if (!tdn) {
			fprintf(stderr, "dn not found\n");
			Debug (LDAP_DEBUG_ANY, "dn not found\n", 0, 0, 0);
			free(obj);
			continue;
		}
		dn = ldif_apply_env_vars(tdn);
		if (tdn) free(tdn);

		objclass_no = MAX_OBJCLASS_NO;
		if ((objclass = (char **) calloc(objclass_no, sizeof(char *))) == NULL) {
			fprintf( stderr, "memory allocation failed\n" );
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			free(obj);
			free(dn);
			fclose(fptr);
			return -1;
		}

		activeobj = 0;
		activesch = 0;
		obj_no = 0;
		objclass_inx = 0;
		tmp = ldif_get_attribute(obj, "objectclass:", &objclass_inx);
		while( tmp ) {
			if (obj_no >= objclass_no) {
				objclass_no += MAX_OBJCLASS_NO;
				if ((objclass = (char **) realloc(objclass, objclass_no * sizeof(char *))) == NULL) {
						fprintf(stderr, "memory allocation failed\n");
						Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
						free(obj);
						free(dn);
						(void) ldif_free_strlist(objclass, obj_no);
						free(objclass);
						fclose(fptr);
						return -1;
				}
			}
			objclass[obj_no++] = tmp;
			if (!activeobj && !strcasecmp(tmp, "GlobusActiveObject"))
				activeobj = 1;
			if (!activesch && !strcasecmp(tmp, "GlobusActiveSearch"))
				activesch = 1;
			tmp = ldif_get_attribute(obj, "objectclass:", &objclass_inx);
		}
		if (obj_no >= objclass_no) {
			objclass_no += 1;
			if ((objclass = (char **) realloc(objclass, objclass_no * sizeof(char *))) == NULL) {
					fprintf(stderr, "memory allocation failed\n");
					Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
					free(obj);
					free(dn);
					(void) ldif_free_strlist(objclass, obj_no);
					free(objclass);
					fclose(fptr);
					return -1;
			}
		}
		objclass[obj_no] = NULL;

		typestr = ldif_get_attribute(obj, "type:", NULL);
		if (!typestr && activeobj) {
			fprintf(stderr, "object type not found\n");
			Debug (LDAP_DEBUG_ANY, "object type not found\n", 0, 0, 0);
			free(dn);
			free(obj);
			(void) ldif_free_strlist(objclass, obj_no);
			free(objclass);
			continue;
		}
		if (activeobj) {
			if ( !strcasecmp(objtypes[EXEC], typestr )) 
				type = EXEC;
			else if ( !strcasecmp(objtypes[DYNAMIC], typestr ))
				type = DYNAMIC;
			else if ( !strcasecmp(objtypes[STATIC], typestr ))
				type = STATIC;
			else {
				fprintf(stderr, "Invalid object type: %s\n", typestr);
				Debug (LDAP_DEBUG_ANY, "Invalid object type: %s\n", typestr, 0, 0);
				free(dn);
				free(obj);
				(void) ldif_free_strlist(objclass, obj_no);
				free(objclass);
				free(typestr);
				continue;
			}
			free (typestr);
		}
		else {
			type = INVALID;
			if(typestr) free(typestr);
		}

		base = ldif_get_attribute(obj, "base:", NULL);
		if (!base && activeobj) {
			fprintf(stderr, "base not found\n");
			Debug (LDAP_DEBUG_ANY, "base not found\n", 0, 0, 0);
			free(dn);
			free(obj);
			(void) ldif_free_strlist(objclass, obj_no);
			free(objclass);
			continue;
		}

		tpath = ldif_get_attribute(obj, "path:", NULL);
		path = ldif_apply_env_vars(tpath);
		if (tpath) free(tpath);
		targs = ldif_get_attribute(obj, "args:", NULL);
		args = ldif_apply_env_vars(targs);
		if (targs) free(targs);
		cachetimestr = ldif_get_attribute(obj, "cachetime:", NULL);
		timelimitstr = ldif_get_attribute(obj, "timelimit:", NULL);
		sizelimitstr = ldif_get_attribute(obj, "sizelimit:", NULL);

		if (cachetimestr) {
			errno = 0;
			cachetime = atoi(cachetimestr);
			if (errno) {
				fprintf(stderr, "Invalid cachetime: %s\n", cachetimestr);
				Debug (LDAP_DEBUG_ANY, "Invalid cachetime\n", cachetimestr, 0, 0);
				free(dn);
				free(obj);
				(void) ldif_free_strlist(objclass, obj_no);
				free(objclass);
				if(base) free(base);
				if(path) free(path);
				if(args) free(args);
				if(cachetimestr) free(cachetimestr);
				if(timelimitstr) free(timelimitstr);
				if(sizelimitstr) free(sizelimitstr);
				continue;
			}
			free(cachetimestr);
		}
		else if (!activesch)
			cachetime = 0;
		else {
			fprintf(stderr, "cachetime not found\n");
			Debug (LDAP_DEBUG_ANY, "cachetime not found\n", 0, 0, 0);
			free(dn);
			free(obj);
			(void) ldif_free_strlist(objclass, obj_no);
			free(objclass);
			if(base) free(base);
			if(path) free(path);
			if(args) free(args);
			if(timelimitstr) free(timelimitstr);
			if(sizelimitstr) free(sizelimitstr);
			continue;
		}

		if (timelimitstr) {
			errno = 0;
			timelimit = atoi(timelimitstr);
			if (errno) {
				fprintf(stderr, "Invalid timelimit: %s\n", timelimitstr);
				Debug (LDAP_DEBUG_ANY, "Invalid timelimit: %s\n", timelimitstr, 0, 0);
				free(dn);
				free(obj);
				(void) ldif_free_strlist(objclass, obj_no);
				free(objclass);
				if(base) free(base);
				if(path) free(path);
				if(args) free(args);
				free (timelimitstr);
				if(sizelimitstr) free(sizelimitstr);
				continue;
			}
			free(timelimitstr);
		}
		else
			timelimit = 0;

		if (sizelimitstr) {
			errno = 0;
			sizelimit = atoi(sizelimitstr);
			if (errno) {
				fprintf(stderr, "Invalid sizelimit: %s\n", sizelimitstr);
				Debug (LDAP_DEBUG_ANY, "Invalid sizelimit: %s\n", sizelimitstr, 0, 0);
				free(dn);
				free(obj);
				(void) ldif_free_strlist(objclass, obj_no);
				free(objclass);
				if(base) free(base);
				if(path) free(path);
				if(args) free(args);
				free (sizelimitstr);
				continue;
			}
			free(sizelimitstr);
		}
		else
			sizelimit = 0;

		ldap_pvt_thread_mutex_lock(&ldt->glist_mutex);
		if ( ldt->total_inx >= data_size ) {
			data_size += LDIF_DATA_SIZE;
			*glist = (ActiveObj **) realloc(*glist, data_size * sizeof(ActiveObj *));
			if (!*glist) {
				fprintf( stderr, "memory allocation failed\n" );
				Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
				free(dn);
				free(obj);
				(void) ldif_free_strlist(objclass, obj_no);
				free(objclass);
				if(base) free(base);
				if(path) free(path);
				if(args) free(args);
				ldap_pvt_thread_mutex_unlock(&ldt->glist_mutex);
				fclose(fptr);
				return -1;
			}
		}

		(*glist)[ldt->total_inx] = (ActiveObj *) calloc( 1, sizeof(ActiveObj));
		if (!(*glist)[ldt->total_inx]) {
			fprintf( stderr, "memory allocation failed\n" );
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			free(dn);
			free(obj);
			(void) ldif_free_strlist(objclass, obj_no);
			free(objclass);
			if(base) free(base);
			if(path) free(path);
			if(args) free(args);
			ldap_pvt_thread_mutex_unlock(&ldt->glist_mutex);
			fclose(fptr);
			return -1;
		}
		((*glist)[ldt->total_inx])->current = 0;
		((*glist)[ldt->total_inx])->mod = 2;
		((*glist)[ldt->total_inx])->dn = dn;
		((*glist)[ldt->total_inx])->objectclass = objclass;
		((*glist)[ldt->total_inx])->base = base;
		((*glist)[ldt->total_inx])->type = type;
		((*glist)[ldt->total_inx])->path = path;
		((*glist)[ldt->total_inx])->args = ldif_get_args(path, base, args);
		((*glist)[ldt->total_inx])->cachetime = cachetime;
		((*glist)[ldt->total_inx])->timelimit = timelimit;
		((*glist)[ldt->total_inx])->sizelimit = sizelimit;
		((*glist)[ldt->total_inx])->tlimit = 0;
		((*glist)[ldt->total_inx])->slimit = 0;
		((*glist)[ldt->total_inx])->stoptime = 0;
		((*glist)[ldt->total_inx])->dataobj_gris = (DataObj_gris **) calloc (((*glist)[ldt->total_inx])->mod, sizeof(DataObj_gris *));
		if(!((*glist)[ldt->total_inx])->dataobj_gris){
		        fprintf( stderr, "memory allocation failed\n" );
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			free(dn);
			free(obj);
			(void) ldif_free_strlist(objclass, obj_no);
			free(objclass);
			if(base) free(base);
			if(path) free(path);
			if(args) free(args);
			ldap_pvt_thread_mutex_unlock(&ldt->glist_mutex);
			fclose(fptr);
			return -1;
		}
		if (!activeobj) {
		        e = str2entry(ldif_strip_comments(obj));
		        ((*glist)[ldt->total_inx])->dataobj_gris[((*glist)[ldt->total_inx])->current]->e = e;
			((*glist)[ldt->total_inx])->dataobj_gris[((*glist)[ldt->total_inx])->current]->next = NULL;
			((*glist)[ldt->total_inx])->lastmodified = ldif_current_time();
		}
		else {
			int jj;
			for(jj = 0; jj < ((*glist)[ldt->total_inx])->mod; jj++)
		        	((*glist)[ldt->total_inx])->dataobj_gris[jj] = NULL;
			((*glist)[ldt->total_inx])->lastmodified = 0;
		}

		((*glist)[ldt->total_inx])->activeobject = activeobj;
		((*glist)[ldt->total_inx])->activesearch = activesch;
		ldap_pvt_thread_mutex_init( &((*glist)[ldt->total_inx])->active_obj_mutex);

		if (args) free(args);
		if (obj) free(obj);

		ldt->total_inx++;
		ldap_pvt_thread_mutex_unlock(&ldt->glist_mutex);
	}

	fclose(fptr);

	return( 0 );
}

char *
ldif_getstrent(FILE *rfp)
{
	char *line, *res = NULL, *obj = NULL, buffer[BUFF_SIZE];
	int len, cur = 0, max = BUFF_SIZE;

	obj = (char *) calloc( max, sizeof(char) );
	(void) memset(obj, 0, max);

	while ( (res = fgets(buffer, sizeof(buffer), rfp)) != NULL) {

		line = ldif_strip_comments(buffer);
		if (!line)
			return NULL;

		len = strlen( line );
		if (len <= 0) {
			if (line) free(line);
			continue;
		}

		if (line[0] != '\n') {
			while ( cur + len + 1 > max ) { 
				max += BUFF_SIZE; 
				obj = (char *) realloc( obj, max );
			}
			strcpy( obj + cur, line );
			cur += len;
			buffer[0] = '\0';
			if (line) free(line);
		}
		else {
			buffer[0] = '\0';
			if (line) free(line);
			if (obj[0])
				return obj;
			else
				continue;
		}
	}
	buffer[0] = '\0';

	if (obj[0]) return obj;
	else {
		if (obj) free(obj);
		return NULL;
	}
}

char *
ldif_getdatastr(FILE *rfp)
{
	char *res = NULL, *obj = NULL, buffer[BUFF_SIZE];
	int len, cur = 0, max = BUFF_SIZE;

	obj = (char *) calloc( max, sizeof(char) );
	(void) memset(obj, 0, max);

	while ( (res = fgets(buffer, sizeof(buffer), rfp)) != NULL) {

		len = strlen( buffer );
		while ( cur + len + 1 > max ) { 
			max += BUFF_SIZE; 
			obj = (char *) realloc( obj, max );
		}
		strcpy( obj + cur, buffer );
		cur += len;

		buffer[0] = '\0';
	}
	buffer[0] = '\0';

	if (obj[0])
		return obj;
	else {
		free(obj);
		return NULL;
	}
}


int ldif_current_time()
{
	return (int) time(NULL);
}


char *ldif_get_attribute(char *str, char *pattern, int *index)
{
	int	no, inx, length;
	char	*tmpstr, *resstr, *teststr, *result;

	inx = 0;

	if (!str || !pattern) {
		fprintf(stderr, "NULL String\n");
		return NULL;
	}

	if (inx >= strlen(str))
		return NULL;

	if (index)
		teststr = str + *index;
	else
		teststr = str;

	if ( (resstr = strstr(teststr, pattern)) == NULL) {
		return NULL;
	}

	/* deal with non-blank chars */
	tmpstr = resstr;
	while(tmpstr > teststr) {
		if (*--tmpstr == '\n')
			break;
		if (*tmpstr != ' ' && *tmpstr != '\t') {

			teststr = resstr + strlen(pattern);

			if ( (resstr = strstr(teststr, pattern)) == NULL) {
				return NULL;
			}
			tmpstr = resstr;
		}
	}

	length = strlen(resstr);
	if (length <= 1) {
		return NULL;
	}

	if ( (result = (char *) calloc(length, sizeof(char))) == NULL) {
		fprintf(stderr, "memory allocation error\n");
		Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
		return NULL;
	}

	while(inx < length && resstr[inx] != ':') {
		inx++;
	}
	inx++;

	while(inx < length && resstr[inx] == ' ') {
		inx++;
	}

	if (inx >= length) {
		if (result) free(result);
		return NULL;
	}

	no = 0;
	while(inx < length) {
		if (resstr[inx] != '\n') {
			result[no++] = resstr[inx++];
		}
		else {
			result[no] = 0;

			/* get rid of spaces at the end of strings */
			while (no > 0) {
				if(result[no-1] == ' ' || result[no-1] == '\t') {
					result[no-1] = 0;
					no--;
				}
				else
					break;
			}
			if (index) *index = resstr - str + inx;
			return result;
		}
	}

	if (result) free(result);
	return NULL;
}


int ldif_clear_list(ActiveObj **list, int total_inx)
{
ActiveObj *current, *tmp;
DataObj_gris *cdataobj_gris, *ndataobj_gris; 
int inx=0;

	int no, i;

	if (!list) return 0;

	while(inx < total_inx) {
		if (!(list[inx])) {
			inx++;
			continue;
		}
		ldap_pvt_thread_mutex_lock(&(list[inx])->active_obj_mutex);
		if ((list[inx])->dn) free((list[inx])->dn);

		no = 0;
		if ((list[inx])->objectclass)
			while((list[inx])->objectclass[no]) {
				free ((list[inx])->objectclass[no]);
				no++;
			}
		if ((list[inx])->objectclass) free((list[inx])->objectclass);

		if ((list[inx])->base) free((list[inx])->base);
		if ((list[inx])->path) free((list[inx])->path);

		no = 0;
		if ((list[inx])->args)
			while((list[inx])->args[no]) {
				free ((list[inx])->args[no]);
				no++;
			}
		if ((list[inx])->args) free((list[inx])->args);

		if ((list[inx])->dataobj_gris){
		     for(i=0; i<(list[inx])->mod; i++){
		          if(list[inx]->dataobj_gris[i]){
			       cdataobj_gris = list[inx]->dataobj_gris[i];
			       ndataobj_gris = cdataobj_gris;
			       while(ndataobj_gris){
				    ndataobj_gris = cdataobj_gris->next;
				    entry_free(cdataobj_gris->e);
				    free(cdataobj_gris);
				    cdataobj_gris = ndataobj_gris;
			       }
			  }
		     }
		}
		ldap_pvt_thread_mutex_unlock(&(list[inx])->active_obj_mutex);
		ldap_pvt_thread_mutex_destroy(&(list[inx])->active_obj_mutex);
		if (list[inx]) free(list[inx]);
		inx++;
	}
	return 0;
}


char **ldif_get_args(char *path, char *base, char *args)
{

	int inx, leng, cnt, arg_no = MAX_ARGS_NO, lpath, lbase; 
	int size, no;
	char **argv, *str, *res, tmpc, *cmd;

	if (!base) return NULL;

	if ((argv = (char **) calloc(arg_no, sizeof(char *))) == NULL) {
		fprintf(stderr, "memory allocation failed\n");
		Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
		return NULL;
	}

	if (!path) {
		if ((cmd = (char *) strdup(base)) == NULL) {
			fprintf(stderr, "memory allocation failed\n");
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			free(argv);
			return NULL;
		}
	}
	else {
		lpath = strlen(path);
		lbase = strlen(base);

		if ( path[lpath-1] == '/' && base[0] == '/') {
			if ((cmd = (char *) calloc(lpath+lbase+2, sizeof(char))) == NULL) {
				fprintf(stderr, "memory allocation failed\n");
				Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
				free(argv);
				return NULL;
			}
			strcpy(cmd, path);
			strcat(cmd, ".");	
			strcat(cmd, base);
		}
		else if ( path[lpath-1] == '/' || base[0] == '/') {
			if ((cmd = (char *) calloc(lpath+lbase+1, sizeof(char))) == NULL) {
				fprintf(stderr, "memory allocation failed\n");
				Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
				free(argv);
				return NULL;
			}
			strcpy(cmd, path);
			strcat(cmd, base);
		}
		else {
			if ((cmd = (char *) calloc(lpath+lbase+2, sizeof(char))) == NULL) {
				fprintf(stderr, "memory allocation failed\n");
				Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
				free(argv);
				return NULL;
			}
			strcpy(cmd, path);
			strcat(cmd, "/");	
			strcat(cmd, base);
		}
	}

	argv[0] = cmd;

	if (!args) {
		argv[1] = NULL;
		return argv;
	}

	leng = strlen(args);

	inx = 0;
	while(inx < leng) 
		if (args[inx] == ' ') 
			inx++;
		else
			break;

	if (inx >= leng) { 
		argv[1] = NULL;
		return argv;
	}

	cnt = 1;
	while (inx < leng) {
		if ( (str = strchr(args + inx, ' ')) != NULL) {
			size = str - args - inx + 1;
			if ((res = (char *) calloc(size, sizeof(char))) == NULL) {
				fprintf(stderr, "memory allocation failed\n");
				Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
				(void) ldif_free_strlist(argv, cnt-1);
				free(argv);
				str[0] = tmpc;
				return NULL;
			}
			for (no = 0; no < size-1; no++)
				res[no] = args[inx + no];
			res[no] = 0;

			if (strlen(res)) {
				if (cnt >= arg_no) {
					arg_no += MAX_ARGS_NO;
					if ((argv = (char **) realloc(argv, arg_no * sizeof(char *))) == NULL) {
						fprintf(stderr, "memory allocation failed\n");
						Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
						(void) ldif_free_strlist(argv, cnt);
						free(argv);
						free(res);
						return NULL;
					}
				}
				argv[cnt++] = res;
			}
			else
				if (res) free(res);

			inx = (int) str - (int) args + 1;
		}
		else
			break;
	}

	if ((res = (char *) strdup(args+inx)) == NULL) {
		fprintf(stderr, "memory allocation failed\n");
		Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
		(void) ldif_free_strlist(argv, cnt);
		free(argv);
		return NULL;
	}

	if (strlen(res)) {
		if (cnt >= arg_no) {
			arg_no += 2;
			if ((argv = (char **) realloc(argv, arg_no * sizeof(char*))) == NULL) {
				fprintf(stderr, "memory allocation failed\n");
				Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
				(void) ldif_free_strlist(argv, cnt);
				free(argv);
				return NULL;
			}
		}
		argv[cnt++] = res;
	}
	if (cnt >= arg_no) {
		arg_no += 1;
		if ((argv = (char **) realloc(argv, arg_no * sizeof(char *))) == NULL) {
			fprintf(stderr, "memory allocation failed\n");
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			(void) ldif_free_strlist(argv, cnt);
			free(argv);
			return NULL;
		}
	}
	argv[cnt] = NULL;

	return argv;
}

int ldif_free_strlist(char **argv, int no)
{

	int inx = 0;

	while(inx < no) {
		if (argv[inx]) free(argv[inx]);
		inx++;
	}

	return 0;
}


char * ldif_strip_comments(char *obj)
{

	int i, no, size, cnt, add_nl;
	char *res, *spinx, *nlinx, *tmp;

	if (!obj) return NULL;

	spinx = strchr(obj, '#');
	if (!spinx) {
		if ((res = (char *) strdup(obj)) == NULL) {
			fprintf(stderr, "memory allocation failure\n");
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			return NULL;
		}
		return res;
	}

	size = strlen(obj);
	if ( (res = (char *) calloc(size+1, sizeof(char))) == NULL ) {
		fprintf(stderr, "memory allocation failure\n");
		Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
		return NULL;
	}

	tmp = spinx - 1;
	while (tmp >= obj && (*tmp == ' ' || *tmp == '\t'))
		tmp--;
	if (tmp >= obj && *tmp != '\n')
		add_nl = 1;
	else
		add_nl = 0;

	no = 0;
	cnt = 0;

	while ( spinx ) {
		for (i = no; i < spinx - obj; i++)
			res[cnt++] = obj[i];
		if (add_nl) {
			res[cnt++] = '\n';
		}

		if ((nlinx = strchr(spinx, '\n')) == NULL) {
			spinx = NULL;
			no = size - 1;
			continue;
		}

		no = nlinx - obj + 1;
		spinx = strchr(nlinx, '#');

		if (spinx) {
			tmp = spinx - 1;
			while (tmp >= obj && (*tmp == ' ' || *tmp == '\t'))
				tmp--;
			if (tmp >= obj && *tmp != '\n')
				add_nl = 1;
			else
				add_nl = 0;
		}
	}

	for (i = no; i < size; i++)
		res[cnt++] = obj[i];
	res[cnt] = 0;

	return res;
}

char *
ldif_str_realloc(char *result, int no, int *mf)
{

	while (no >= ENV_VAR_SIZE*(*mf)) {
		*mf += 1;
		result = (char *) realloc(result, ENV_VAR_SIZE*(*mf));
		if (!result) {
			fprintf(stderr, "memory allocation error\n");
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			return NULL;
		}
	}
	return result;
}

char *
ldif_apply_env_vars(char *tstr)
{
	int no, size, i, j, mf_res, mf_str;
	int dflag = 0, oflag = 0;
	int start, end;
	char *str, *result, *environ;
	struct stat buf;

	if (!tstr) {
		fprintf(stderr, "Null String\n");
		return NULL;
	}
	size = strlen(tstr);

	result = (char *) calloc(ENV_VAR_SIZE, sizeof(char));
	if(!result) {
		fprintf(stderr, "memory allocation error\n");
		Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
		return NULL;
	}

	str = (char *) calloc(ENV_VAR_SIZE, sizeof(char));
	if(!str) {
		fprintf(stderr, "memory allocation error\n");
		Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
		if (result) free(result);
		return NULL;
	}

	mf_res = 1;
	mf_str = 1;
	no = 0;
	for (i=0; i<size; i++) {
		switch (tstr[i]) {
			case '$':
				dflag = 1;
				break;
			case '{':
				if (dflag && i < size-1) {
					oflag = 1;
					start = i+1;
				}
				else {
					if ((result = ldif_str_realloc(result, no, &mf_res)) != NULL) {
						result[no++] = tstr[i];
					}
					else {
						if (str) free(str);
						return NULL;
					}
				}
				dflag = 0;
				break;
			case '}':
				if (oflag) {
					end = i;
					for (j=0; j<end-start; j++) {
						if ((str = ldif_str_realloc(str, j, &mf_str)) != NULL) {
							str[j] = tstr[start+j];
						}
						else {
							if (result) free(result);
							return NULL;
						}
					}
					if (j>0) {
						if ((str = ldif_str_realloc(str, j, &mf_str)) != NULL) {
							str[j] = 0;
						}
						else {
							if (result) free(result);
							return NULL;
						}
					}
					else {
						if (str) free(str);
						if (result) free(result);
						return NULL;
					}

					if((environ = getenv(str)) != NULL) {
						for (j=0; j<strlen(environ); j++) {
							if ((result = ldif_str_realloc(result, no, &mf_res)) != NULL) {
								result[no++] = environ[j];
							}
							else {
								if (str) free(str);
								return NULL;
							}
						}
					}
					else {
						fprintf(stderr, "environment \"%s\" not set\n", str);
						Debug(LDAP_DEBUG_ANY, "environment \"%s\" not set\n", str, 0, 0 );
						if (str) free(str);
						if (result) free(result);
						return NULL;
					}
					dflag = 0;
					oflag = 0;
				}
				else {
					if ((result = ldif_str_realloc(result, no, &mf_res)) != NULL) {
						result[no++] = tstr[i];
					}
					else {
						if (str) free(str);
						return NULL;
					}
				}
				dflag = 0;
				break;
			default:
				if(!oflag) {
					if ((result = ldif_str_realloc(result, no, &mf_res)) != NULL) {
						result[no++] = tstr[i];
					}
					else {
						if (str) free(str);
						return NULL;
					}
				}
				dflag = 0;
				break;
		}
	}

	if (oflag && start > 1) {
		for(j=start-2; j<size; j++) {
			if ((result = ldif_str_realloc(result, no, &mf_res)) != NULL) {
				result[no++] = tstr[j];
			}
			else {
				if (str) free(str);
				return NULL;
			}
		}
	}

	if ((result = ldif_str_realloc(result, no, &mf_res)) != NULL) {
		result[no] = 0;
	}
	else {
		if (str) free(str);
		return NULL;
	}

	if (str) free(str);

	return result;
}

int ldif_get_local_port()
{
	int port;
	char *arg = NULL, *pstr1 = NULL, *pstr2 = NULL;
	FILE *fptr;

	if ( slapd_args_file == NULL ) {
		fprintf( stderr, "slapd_args_file not found\n" );
		Debug( LDAP_DEBUG_ANY, "slapd_args_file not found\n", 0, 0, 0 );
		return( -1 );
        }

	if ( (fptr = fopen( slapd_args_file, "r" )) == NULL ) {
		fprintf (stderr, "Can't open %s\n", slapd_args_file);
		Debug (LDAP_DEBUG_ANY, "Can't open %s\n", slapd_args_file, 0, 0 );
		return -1;
	}

	arg = ldif_getstrent(fptr);
	if (!arg) {
		fprintf (stderr, "Invalid args file (%s)\n", slapd_args_file);
		Debug (LDAP_DEBUG_ANY, "Invalid args file (%s)\n", slapd_args_file, 0, 0 );
		fclose(fptr);
		return -1;
	}
	fclose(fptr);

	pstr1 = strrchr(arg, ':');
	if (!pstr1) {
		fprintf (stderr, "Invalid args file (%s)\n", slapd_args_file);
		Debug (LDAP_DEBUG_ANY, "Invalid args file (%s)\n", slapd_args_file, 0, 0 );
		free(arg);
		return -1;
	}
	pstr2 = strrchr(arg, ' ');
	if (!pstr2) {
		fprintf (stderr, "Invalid args file (%s)\n", slapd_args_file);
		Debug (LDAP_DEBUG_ANY, "Invalid args file (%s)\n", slapd_args_file, 0, 0 );
		free(arg);
		return -1;
	}
	*pstr2 = '\0';

	errno = 0;
	port = atoi(pstr1+1);
	if (errno != 0) {
		fprintf (stderr, "Invalid args file (%s)\n", slapd_args_file);
		Debug (LDAP_DEBUG_ANY, "Invalid args file (%s)\n", slapd_args_file, 0, 0 );
		free(arg);
		return -1;
	}

	free(arg);

	return port;
}

int ldif_local_process(char* reghost, int regport, int myport)
{

	struct hostent *host_ptr = NULL;
	char reg_hostname[MAX_HOSTNAME], my_hostname[MAX_HOSTNAME];
	char myhost[MAX_HOSTNAME];

	/* 
	   if the giis' port number can't be found, consider it
           as the same port number as the registrant
        */
	if (myport != -1 && regport != myport)
		return 0;

	if( gethostname(myhost, MAX_HOSTNAME) != 0) {
		return -1;
	}

	host_ptr = gethostbyname(reghost);
	if(host_ptr == NULL)
	{
		return -1;
	}
	host_ptr = gethostbyaddr(*host_ptr->h_addr_list, sizeof(struct in_addr), AF_INET);
	if(host_ptr == NULL)
	{
		return -1;
	}
	strcpy(reg_hostname, host_ptr->h_name);
   
	host_ptr = gethostbyname(myhost);
	if(host_ptr == NULL)
	{
		return -1;
	}
	host_ptr = gethostbyaddr(*host_ptr->h_addr_list, sizeof(struct in_addr), AF_INET);
	if(host_ptr == NULL)
	{
		return -1;
	}
	strcpy(my_hostname, host_ptr->h_name);

	return !strcasecmp(reg_hostname, my_hostname);
}

int ldif_check_peer_name(char *peer_name)
{

	char myhost[MAX_HOSTNAME], my_hostname[MAX_HOSTNAME];
	char peerhoststr[MAX_HOSTNAME], peer_hostname[MAX_HOSTNAME];
	struct hostent *host_ptr = NULL;
	char *pstr1, *pstr2;
	int port;

	if(!peer_name) return 0;
	strcpy(peerhoststr, peer_name);

	pstr1 = strrchr(peerhoststr, '=');
	if (!pstr1) {
		return 0;
	}

	pstr2 = strrchr(pstr1, ':');
	if (!pstr2) {
		return 0;
	}

	*pstr2 = '\0';

	host_ptr = gethostbyname(pstr1+1);
	if(host_ptr == NULL)
	{
		return -1;
	}
	host_ptr = gethostbyaddr(*host_ptr->h_addr_list, sizeof(struct in_addr), AF_INET);
	if(host_ptr == NULL)
	{
		return -1;
	}
	strcpy(peer_hostname, host_ptr->h_name);

	if( gethostname(myhost, MAX_HOSTNAME) != 0) {
		return -1;
	}
	host_ptr = gethostbyname(myhost);
	if(host_ptr == NULL)
	{
		return -1;
	}
	host_ptr = gethostbyaddr(*host_ptr->h_addr_list, sizeof(struct in_addr), AF_INET);
	if(host_ptr == NULL)
	{
		return -1;
	}
	strcpy(my_hostname, host_ptr->h_name);
	return !strcasecmp(peer_hostname, my_hostname);
}


