/* search.c - ldif backend search function */
/*
$Header: /home/globdev/CVS/globus-packages/mds/openldap/ldapmodules/search.c,v 1.10.4.1 2003/07/02 23:33:00 olomu Exp $
$Log: search.c,v $
Revision 1.10.4.1  2003/07/02 23:33:00  olomu
modifying previous bugfix bugzilla #956 to check for interrupts in RH 7.3 and above

Revision 1.12  2003/06/12 21:17:39  olomu
commited fix for bugzilla #956.  Changes for select behavior in RH 7.3 and up, a randomly returned -1 for select causes gris pipes to close.

Revision 1.11  2003/05/07 20:09:13  olomu
GRIS Update.  Using linked structure of Openldap Entries for caching GRIS data

Revision 1.10  2003/03/27 03:08:46  jslee
allowing multiple grises in one slapd

Revision 1.9  2003/03/22 02:49:56  jslee
adding allowstaledata yes/no

Revision 1.8  2003/02/05 09:10:41  jslee
fixing threads runout - bugzilla #644

Revision 1.6.2.1  2003/01/10 21:59:03  jslee
removing config file read routine

Revision 1.6  2002/11/20 23:30:45  jslee
fixing memory leaks in case of sizelimit reached

Revision 1.5  2002/10/16 01:18:03  jslee
fixing file descriptor leaks

Revision 1.4  2002/08/14 22:40:30  benc
Merge MDS2.2 into trunk.

Revision 1.3.20.1  2002/08/07 03:56:06  benc
Merged (hopefully) all of the mds_2_2_branch work onto this new branch.

Compiles and runs on Linux.

Revision 1.3.2.4  2002/06/26 20:42:14  jslee
adding a Debug message in case IP's timelimit exceeded

Revision 1.3.2.3  2002/06/11 21:34:09  jslee
changing the first par of select

Revision 1.3.2.2  2002/06/04 00:59:40  jslee
fixing info-provider hanging problems

Revision 1.3.2.1  2002/05/20 21:44:36  jslee
adding syslog messages to the ldif backend

Revision 1.1.1.1.2.1  2002/02/15 04:42:52  jslee
debugging internal node generation

Revision 1.1.1.1  2002/01/10 02:10:09  jslee
importing globus-beta-branch to the main trunk

Revision 1.2.4.7  2001/10/30 01:26:27  jslee
fixing concurrency problems

Revision 1.2.4.6  2001/10/23 02:00:35  jslee
changing the symantics of sizelimit:
 sizelimit of 0 or less means no sizelimit

Revision 1.2.4.5  2001/10/13 03:45:33  jslee
adding GlobusStub object as default

Revision 1.2.4.4  2001/10/11 21:24:09  jslee
fixing timelimit, sizelimit routines

Revision 1.2.4.3  2001/09/28 07:23:17  jslee
resolving naming conflicts with giis backend routines

Revision 1.2.4.2  2001/09/26 21:18:38  smartin
committing merge from main repository trunk to beta branch

Revision 1.3  2001/09/24 15:13:10  smartin
committing merge from globus-beta-branch

Revision 1.2.4.1  2001/09/21 19:58:35  meisu
1) rename ldif.h to slapd_ldif.h
2) alter the Makefile.am to match with giis-ldapmodule's and the
   flavor is in the install path

Revision 1.2  2001/09/19 20:37:53  meisu
1) upgrading 2.0.14 and some cleanup and bug fixes

Revision 1.1  2001/08/06 22:15:39  bletzing
New Package.  Doesn't Build.

Revision 1.1.1.1  2001/07/06 21:11:15  mei
1) splitting out back-ldif from the openldap 2.0.8-patched into a standalone
   directory that can generate 'dynamic modules' that can be used by the
   openldap's slapd (checked against 2.0.11 already)

Revision 1.8  2001/05/30 22:07:31  june
fixing read_and_send_results bug

Revision 1.7  2001/05/24 03:04:24  june
remove debug printfs

Revision 1.6  2001/05/24 01:05:06  june
gridmap checking & timelimit debugging

Revision 1.5  2001/04/20 03:34:21  june
fixing giis timing bug

Revision 1.4  2001/04/19 22:11:35  june
fixing first object disappearing bug

Revision 1.3  2001/04/19 17:54:23  june
updating dispatching routines

Revision 1.2  2001/04/02 20:18:40  june
debugging bind_dn

Revision 1.1  2001/03/24 00:08:38  june
adding ldif backend

Revision 1.1  2001/03/22 03:01:25  june
add ldif backend

Revision 1.5  2000/10/03 23:11:56  june
add bind_dn

Revision 1.4  2000/09/23 00:06:06  june
add forkandexec

Revision 1.3  2000/09/13 02:11:50  june
resolve changes between 2.0 & 2.0.1

Revision 1.2  2000/09/12 22:41:39  june
Add ldif backend
*/


#include "portable.h"

#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>

#include <ac/ctype.h>
#include <ac/socket.h>
#include <ac/string.h>
#include <ac/time.h>

#include "slap.h"
#include "slapd_ldif.h"
#include <ldap_pvt.h>

int
ldif_back_search(
    Backend	*be,
    Connection	*conn,
    Operation	*op,
    const char	*base,
    const char	*nbase,
    int		scope,
    int		deref,
    int		slimit,
    int		tlimit,
    Filter	*filter,
    const char	*filterstr,
    char	**attrs,
    int		attrsonly
)
{
	Entry		*e;
	char		*s;
	time_t		stoptime;

	int sent = 0, i;
	int err = LDAP_SUCCESS;

	char *str = NULL;
	char *cmdstr = NULL;
	char *cmdstr1 = NULL;
	char *cmdstr2 = NULL;
	char *binddn = NULL;

	pid_t pid;
	pid_t cpid;
	int status;

	FILE *rfp;
	FILE *wfp;

	ActiveObj	*current = NULL;
	DataObj_gris    *Dataobj_gris=NULL, *cdataobj_gris=NULL, *ndataobj_gris=NULL, *sdataobj_gris=NULL, *edataobj_gris=NULL, *tempdataobj_gris;
	int		ttl, create_new, inx, sfxm, fdn, cntrl;
	FILE		*fptr;
	char		*dn, *data, *obj, *path, *sfx, *ttlstr, *tempdata, *data1;
	int		cur_inx, errno;
	char		*authcid, *ubuf = NULL;
	int		search_timed_out;

	AttributeDescription *ad_objectClass = slap_schema.si_ad_objectClass;
	LdifDT *ldt = (LdifDT *) be->be_private;

        /* LdifDT field copies */
	int             ldifallow_anonym;
	char            *ldifconfigfile;
	ActiveObj       **glist;
	int             total_inx;
	int             local_port;
	int             allow_stale_data;
	char            *ldifgiven_sfx;
	Entry		*stub_entry;

	authcid = (char *) strchr(op->o_dn, '=');
	if ( authcid && strlen(authcid) >= 1 ) 
		authcid += 1;

        if ( ldt == NULL ) {
                send_ldap_result( conn, op, LDAP_UNWILLING_TO_PERFORM, NULL,
                    "search not implemented", NULL, NULL );
                Debug( LDAP_DEBUG_ANY, "search not implemented\n", 0, 0, 0 );
                return( -1 );
        }

        /* initialize LdifDT fields locally */
        ldap_pvt_thread_mutex_lock(&ldt->glist_mutex);
        ldifallow_anonym = ldt->ldifallow_anonym;
        total_inx = ldt->total_inx;
        glist = *ldt->glist;
        ldifconfigfile = ldt->ldifconfigfile;
        ldifgiven_sfx = ldt->ldifgiven_sfx;
        local_port = ldt->local_port;
        allow_stale_data = ldt->allow_stale_data;
	stub_entry = ldt->stub_entry;
        ldap_pvt_thread_mutex_unlock(&ldt->glist_mutex);

	/* if the query's from the local machine, skip the gridmap check */
	if (!ldifallow_anonym && ldif_check_peer_name(conn->c_peer_name) != 1) {
	 	if ( authcid == NULL || globus_gss_assist_gridmap(authcid, &ubuf)!=0) {
			send_ldap_result( conn, op, LDAP_UNWILLING_TO_PERFORM, NULL, "access denied: gridmap lookup failed", NULL, NULL );
			if (ubuf) free(ubuf);
			goto done;
		}
		if (ubuf) free(ubuf);
	}

	if ( be->be_private == NULL ) {
		send_ldap_result( conn, op, LDAP_UNWILLING_TO_PERFORM, NULL,
		    "search not implemented", NULL, NULL );
		return( -1 );
        }


	if (!glist) {
		fprintf (stderr, "initialization error\n");
		Debug (LDAP_DEBUG_ANY, "initialization error\n", 0, 0, 0);
		return -1;
	}

	tlimit = (tlimit > be->be_timelimit || tlimit < 1) ? 
			be->be_timelimit : tlimit;
	stoptime = ldif_current_time() + tlimit;
	slimit = (slimit > be->be_sizelimit || slimit < 1) ? 
			be->be_sizelimit : slimit;

	create_new = 1;

	cur_inx = 0;
	inx = 0;

	while (inx < total_inx)
	  {
	        ldap_pvt_thread_mutex_lock(&(glist[inx])->active_obj_mutex);
		
		if ((glist[inx])->timelimit > 0)
			(glist[inx])->tlimit = (glist[inx])->timelimit;
		else
			(glist[inx])->tlimit = tlimit;

		if ((glist[inx])->sizelimit > 0)
			(glist[inx])->slimit = (glist[inx])->sizelimit;
		else
			(glist[inx])->slimit = -1;
		(glist[inx])->stoptime = ldif_current_time() + (glist[inx])->tlimit;

		sfxm = ldif_dispatch_active_obj((glist[inx])->dn, nbase, scope);
		if (sfxm == -1) 
		  {
			ldap_pvt_thread_mutex_unlock(&(glist[inx])->active_obj_mutex);
			return -1;
		  }
		if(sfxm) {
			if (ldif_is_old(glist[inx]) && (glist[inx])->activeobject){
				create_new = 1;
			}
			else {/* CACHE HIT STEP*/
				create_new = 0;

				cur_inx = 0;
				Dataobj_gris = (glist[inx])->dataobj_gris[glist[inx]->current];
				while (Dataobj_gris) 
				  {
					if (!Dataobj_gris->e->e_ndn) 
					  {
                                                Dataobj_gris = Dataobj_gris->next;
						continue;
					  }
					if (ldif_dispatch_obj(Dataobj_gris->e->e_ndn, nbase, scope) != 1)
					  {
						Dataobj_gris = Dataobj_gris->next;
						continue;
					  }

					/* check time limit */
					if ( ldif_current_time() > stoptime ) 
					  {
						send_ldap_result( conn, op, LDAP_TIMELIMIT_EXCEEDED,
				    		NULL, NULL, NULL, NULL );
						ldap_pvt_thread_mutex_unlock(&(glist[inx])->active_obj_mutex);
						goto done;
					  }

					/* check time limit for the entry */
					if ( ldif_current_time() > (glist[inx])->stoptime ) 
					  {
					        Dataobj_gris = Dataobj_gris->next;
						break;
					  }

					if ( Dataobj_gris->e && test_filter( be, conn, op, Dataobj_gris->e, filter ) == LDAP_COMPARE_TRUE ) 
					  {
						/* check size limit */
						if ( --slimit == -1 ) {
							send_ldap_result( conn, op, LDAP_SIZELIMIT_EXCEEDED,
					    		NULL, NULL, NULL, NULL );
							ldap_pvt_thread_mutex_unlock(&(glist[inx])->active_obj_mutex);
							goto done;
						}

						/* check size limit for the entry */
						if ( --(glist[inx])->slimit == -1 ) {
						        Dataobj_gris = Dataobj_gris->next;
							break;
						}

						send_search_entry( be, conn, op, Dataobj_gris->e, attrs, attrsonly, NULL );
						sent++;
					  }
					Dataobj_gris = Dataobj_gris->next;
				  }
			}
			
		}
		
		
		else 
		  {
		      create_new = 0;
		  }
		
		if (create_new) /*CACHE MISS STEP*/
		  {
		    
		        int rc, start_t, end_t, t_limit;
			fd_set readfds;
			struct timeval tv;
			
			if (ldif_is_old(glist[inx]) && (glist[inx])->activeobject)
			  {
				if ( (void *) ldif_forkandexec( (glist[inx])->args, &rfp, &wfp ) == (void *) -1 ) 
				  {
				        fprintf( stderr, "could not fork/exec\n" );
					Debug (LDAP_DEBUG_ANY, "could not fork/exec\n", 0, 0, 0);
					ldap_pvt_thread_mutex_unlock(&(glist[inx])->active_obj_mutex);
					return( -1 );
				  }

				/* write out the request to the search process */
				fprintf( wfp, "SEARCH\n" );
				fprintf( wfp, "msgid: %ld\n", (long) op->o_msgid );
				print_ldif_suffixes( wfp, be );
				fprintf( wfp, "connid: %lu\n", conn->c_connid );
				fprintf( wfp, "base: %s\n", base );
				fprintf( wfp, "scope: %d\n", scope );

				if (conn->c_dn != NULL) 
				  {
					binddn = strchr(conn->c_dn, '=');
					if (binddn && strlen(binddn) > 1)
						fprintf( wfp, "bind_dn: %s\n", binddn+1);
					else
						fprintf( wfp, "bind_dn: NULL\n");
				  }
				else
					fprintf( wfp, "bind_dn: NULL\n");

				fprintf( wfp, "deref: %d\n", deref );
				fprintf( wfp, "sizelimit: %d\n", slimit );
				fprintf( wfp, "timelimit: %d\n", tlimit );
				fprintf( wfp, "filter: %s\n", filterstr );
				fprintf( wfp, "attrsonly: %d\n", attrsonly ? 1 : 0 );
				fprintf( wfp, "attrs:%s", attrs == NULL ? " all" : "" );
				for ( i = 0; attrs != NULL && attrs[i] != NULL; i++ ) 
				  {
					fprintf( wfp, " %s", attrs[i] );
				  }
				fprintf( wfp, "\n" );
				fclose( wfp );

				fdn = fileno(rfp);
				FD_ZERO( &readfds );
				FD_SET( fdn, &readfds );

				start_t = ldif_current_time();
				t_limit = (glist[inx])->timelimit;
				end_t = start_t + t_limit;

				tv.tv_sec = (glist[inx])->timelimit;
				tv.tv_usec = 0;
				
				errno = 0;
				while (errno == EINTR)
				        rc = select (fdn + 1, &readfds, NULL, &readfds, &tv);
				   

				if (rc < errno) 
				  { 
					search_timed_out = 1;
					if ((glist[inx])->base)
						Debug (LDAP_DEBUG_ANY, "information provider %s: timelimit exceeded\n", (glist[inx])->base, 0, 0);
					else
						Debug (LDAP_DEBUG_ANY, "information provider %d: timelimit exceeded\n", inx, 0, 0);
					data = NULL;
				  }
				else 
				  {
					search_timed_out = 0;
					data = ldif_getdatastr(rfp);
				  }
			  

				data1 = ldif_strip_comments(data);
				if (data) free (data);
				fclose (rfp);
				if(allow_stale_data == 1 || search_timed_out == 0) 
				  {
				    (glist[inx])->lastmodified = ldif_current_time();
				  }
			  }
			
			cur_inx = 0;
			tempdata = ldif_getdataent(data1,&cur_inx);
			while(tempdata) /*decomposing data and creating linked list structure for data*/
			  {
			       dn = ldif_get_attribute(tempdata, "dn:", NULL);
			       if(!dn)
				{
				     if(tempdata) free(tempdata);
				     tempdata = ldif_getdataent(data1,&cur_inx);
				     continue;
				}
			    
			       /*CREATING INITIAL ENTRY*/
			       e = str2entry(tempdata);
			       cdataobj_gris = (DataObj_gris*)calloc(1,sizeof(DataObj_gris));
			       cdataobj_gris->e = e;
			       cdataobj_gris->next = NULL;
			       
			       if(!sdataobj_gris) /*INITIAL ENTRY*/
			       {
				    sdataobj_gris = cdataobj_gris;
				    edataobj_gris = cdataobj_gris;
			       }
			       else /*LINKING ENTRIES*/
			       {
				   edataobj_gris->next = cdataobj_gris;
				   edataobj_gris = cdataobj_gris;
			       }
			    
			    
			       if (dn) free(dn);
			       if(tempdata) free(tempdata);
			       tempdata = ldif_getdataent(data1,&cur_inx);
			  }
			
			if(data1) free(data1);
			tempdataobj_gris = sdataobj_gris; /*SETTING UP DATA FROM INITIAL ENTRY FOR DISPATCH*/
			while(tempdataobj_gris) /*data dispatch*/
			 {
			      if (ldif_dispatch_obj(tempdataobj_gris->e->e_ndn, nbase, scope) != 1)
			       {
				   tempdataobj_gris = tempdataobj_gris->next;
				   continue;
			       }
			      
			    
			      /* check time limit */
			      if ( ldif_current_time() > stoptime ) 
			       {
				   send_ldap_result( conn, op, LDAP_TIMELIMIT_EXCEEDED, NULL, NULL, NULL, NULL );
				   ldap_pvt_thread_mutex_unlock(&(glist[inx])->active_obj_mutex);
				   goto done;
			       }
			    
			      /* check time limit for the entry */
			      if ( ldif_current_time() > (glist[inx])->stoptime ) 
			       {
				   tempdataobj_gris = tempdataobj_gris->next;
				   break;
			       } 
			    
			      if ( tempdataobj_gris->e && test_filter( be, conn, op, tempdataobj_gris->e, filter ) == LDAP_COMPARE_TRUE ) 
			      {
				   /* check size limit */
				   if ( --slimit == -1 ) 
				    {
				         send_ldap_result( conn, op, LDAP_SIZELIMIT_EXCEEDED, NULL, NULL, NULL, NULL );
					 ldap_pvt_thread_mutex_unlock(&(glist[inx])->active_obj_mutex);
					 goto done;
				    }
				
				   /* check size limit for the entry */
				   if ( --(glist[inx])->slimit == -1 ) 
				    {
				         tempdataobj_gris = tempdataobj_gris->next;
					 break;
				    }
				
				   send_search_entry( be, conn, op, tempdataobj_gris->e, attrs, attrsonly, NULL );
				   sent++;
				
			      }
			      tempdataobj_gris = tempdataobj_gris->next;

			  }
			
			/*CALCULATE BUFFER TO BE USED TO CACHE DATA*/
			cntrl = (glist[inx])->current;
			cntrl = (cntrl +1) % (glist[inx])->mod;

			/*CHECK FOR AND FREE EXISTING DATA*/
			if((glist[inx])->dataobj_gris[cntrl])
			 {
			      cdataobj_gris = (glist[inx])->dataobj_gris[cntrl];
			      ndataobj_gris = cdataobj_gris;
			      while(ndataobj_gris)
			       {
				    ndataobj_gris = cdataobj_gris->next;
				    entry_free(cdataobj_gris->e);
				    free(cdataobj_gris);
				    cdataobj_gris = ndataobj_gris;
			      }
			  }
			/* CACHING NEW DATA */
			(glist[inx])->dataobj_gris[cntrl] = sdataobj_gris;
			(glist[inx])->current = cntrl;
			sdataobj_gris = NULL;
			edataobj_gris = NULL;
			ndataobj_gris = NULL;
			cdataobj_gris = NULL;
			
		  }
		
		ldap_pvt_thread_mutex_unlock(&(glist[inx])->active_obj_mutex);
		inx++;
	  }
	
 done:
	if ( same_sfx(ldifgiven_sfx, "mds-vo-name=local,o=grid") && stub_entry && test_filter( be, conn, op, stub_entry, filter ) == LDAP_COMPARE_TRUE ) 
	 {
	     dn = (char *) strdup(be->be_suffix[0]);
	     if(!dn) 
	      {
	           fprintf(stderr, "memory allocation error\n");
		   Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
		   return -1;
	      }
	     if (ldif_dispatch_obj(dn, nbase, scope) == 1) 
	      { 
	          send_search_entry( be, conn, op, stub_entry, attrs, attrsonly, NULL ); 
		  sent++;
	      }
	     free(dn);
	 }
	
	send_ldap_result( conn, op, err, NULL, NULL, NULL, NULL );
	
	return( 0 );
}

int same_sfx(char *sfx1, char *sfx2)
{
	char *csfx1=NULL, *csfx2=NULL;
	int result;

	if ((csfx1 = ldif_canonicalize_dn(sfx1)) == NULL)
		return -1;
	if ((csfx2 = ldif_canonicalize_dn(sfx2)) == NULL) {
		free (csfx1);
		return -1;
	}
	if (!strcasecmp(csfx1, csfx2))
		result = 1;
	else
		result = 0;

	if(csfx1) free(csfx1);
	if(csfx2) free(csfx2);
	return result;
}

int ldif_dispatch_obj(char *dn, char *base, int scope)
{

	char *cdn, *cbase, *res;
	int  lcdn, lbase, i, result;

	if ((cdn = ldif_canonicalize_dn(dn)) == NULL)
		return 0;
	if ((cbase = ldif_canonicalize_dn(base)) == NULL) {
		free (cdn);
		return 0;
	}

	/* base == dn */
	if (!strcasecmp(cdn, cbase)) {
		if (scope == 1)
			result = 0;
		else
			result = 1;
	}

	/* base is an ancestor of dn */
	else if ((res = strstr(cdn, cbase)) != NULL && !strcasecmp(res, cbase)) {

		lcdn = strlen(cdn);
		lbase = strlen(res);

		switch (scope) {
		case 0:
			result = 0;
			break;
		case 1:
			result = 1;
			for (i=0; i<lcdn - lbase - 1; i++)
				if (cdn[i] == ',') {
					/* base is not direct parent of dn */
					result = 0;
					break;
				}
			break;
		case 2:
			result = 1;
			break;
		default:
			fprintf(stderr, "invalid scope: %d\n", scope);
			Debug (LDAP_DEBUG_ANY, "invalid scope: %d\n", scope, 0, 0);
			result = -1;
			break;
		}
	}
	/* dn is an ancestor of base or dn & base are disjoint */
	else
		result = 0;

	free(cdn);
	free(cbase);

	return result;
}

int ldif_dispatch_active_obj(char *dn, char *base, int scope)
{

	char *cdn, *cbase, *res;
	int  lcdn, lbase, i, result;

	if ((cdn = ldif_canonicalize_dn(dn)) == NULL)
		return 0;
	if ((cbase = ldif_canonicalize_dn(base)) == NULL) {
		free (cdn);
		return 0;
	}

	/* base == dn */
	if (!strcasecmp(cdn, cbase)) {
		result = 1;
	}

	/* check if dn is an ancestor of base */
	else if ((res = strstr(cbase, cdn)) != NULL && !strcasecmp(res, cdn)) {
		result = 1;
	}

	/* check if base is an ancestor of dn */
	else if ((res = strstr(cdn, cbase)) != NULL && !strcasecmp(res, cbase)) {

		lcdn = strlen(cdn);
		lbase = strlen(cbase);

		switch (scope) {
		case 0:
			result = 0;
			break;
		case 1:
			result = 1;
			for (i=0; i<lcdn - lbase - 1; i++)
				if (cdn[i] == ',') {
					/* base is not direct parent of dn */
					result = 0;
					break;
				}
			break;
		case 2:
			result = 1;
			break;
		default:
			fprintf(stderr, "invalid scope: %d\n", scope);
			Debug (LDAP_DEBUG_ANY, "invalid scope: %d\n", scope, 0, 0);
			result = -1;
			break;
		}
	}
	/* base and dn are disjoint */
	else
		result = 0;

	free(cdn);
	free(cbase);

	return result;
}


int ldif_is_old(ActiveObj *current)
{

	if (current->cachetime <= 0)
		return 1;

	if (ldif_current_time() - current->lastmodified > current->cachetime)
		return 1;
	else
		return 0;
}


char * ldif_canonicalize_dn(char *dn)
{

	int i, no, size, cnt, inx;
	char *tdn, *res, *spinx, *nlinx;

	if (!dn) return NULL;

	tdn = (char *) strdup(dn);
	if (!tdn) {
		fprintf(stderr, "memory allocation failed\n");
		Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
		return NULL;
	}

	size = strlen(tdn);
	for (i=0; i<size; i++)
		tdn[i] = toupper(tdn[i]);

	spinx = strchr(tdn, ',');
	if (!spinx) {
		no = 0;
		while (no < size)
			if (tdn[no] == ' ' || tdn[no] == '\t' || tdn[no] == '\"')
				no++;
			else
				break;
		if (no >= size) {
			free(tdn);
			return NULL;
		}

		if ((res = (char *) strdup(tdn+no)) == NULL) {
			fprintf(stderr, "memory allocation failure\n");
			Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
			free(tdn);
			return NULL;
		}

		inx = strlen(res);
		while(inx > 0)
			if (res[inx-1] == ' ' || res[inx-1] == '\t' || res[inx-1] == '\"')
				res[--inx] = 0;
			else
				break;

		if (inx) {
			free(tdn);
			return res;
		}
		else {
			free(res);
			free(tdn);
			return NULL;
		}
	}

	if ( (res = (char *) calloc(size+1, sizeof(char))) == NULL ) {
		fprintf(stderr, "memory allocation failure\n");
		Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
		free(tdn);
		return NULL;
	}

	no = 0;
	cnt = 0;

	while ( spinx ) {
		while (no < size)
			if (tdn[no] == ' ' || tdn[no] == '\t' || tdn[no] == '\"')
				no++;
			else
				break;

		for (i = no; i <= spinx - tdn; i++)
			res[cnt++] = tdn[i];

		while (cnt > 0)
			if (res[cnt-1] == ' ' || res[cnt-1] == '\t' || res[cnt-1] == '\"')
				cnt--;
			else
				break;

		no = spinx - tdn + 1;
		spinx = strchr(tdn+no, ',');
	}

	while (no < size)
		if (tdn[no] == ' ' || tdn[no] == '\t' || tdn[no] == '\"')
			no++;
		else
			break;

	for (i = no; i < size; i++)
		res[cnt++] = tdn[i];

	while (cnt > 0)
		if (res[cnt-1] == ' ' || res[cnt-1] == '\t' || res[cnt-1] == '\"')
			cnt--;
		else
			break;
	res[cnt] = 0;

	free(tdn);
	return res;
}
char *
ldif_getdataent(char *data, int *cur_inx)
{
	char *res = NULL;
	static ActiveObj *current;
	int no, disp, size, tmp_inx;
	char *inx;

	if (!data)
		return NULL;

	if ((size = strlen(data)) <= 0)
		return NULL;

	if(*cur_inx >= size) 
		return NULL;

	/* get rid of initial blanks */
	if (*(data+*cur_inx) == ' ' || 
	    *(data+*cur_inx) == '\t' ||
	    *(data+*cur_inx) == '\n') 
		(*cur_inx)++;

	if((inx = strstr(data + *cur_inx, "\n\n")) != NULL) {
		disp = (int) (inx - data) - *cur_inx;
		if ((res = (char *) calloc(disp + 1, sizeof(char))) == NULL) {
			fprintf(stderr, "memory allocation failure\n");
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			return NULL;
		}
		for(no=0; no<disp; no++) {
			res[no] = data[*cur_inx + no];
		}
		res[no] = 0;
		*cur_inx = *cur_inx + disp + 2;
		while(*cur_inx < size && data[*cur_inx] == '\n')
				(*cur_inx)++;
		return res;
	}
	else {
		disp = size - *cur_inx;
		if ((res = (char *) calloc(disp + 1, sizeof(char))) == NULL) {
			fprintf(stderr, "memory allocation failure\n");
			Debug (LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0);
			return NULL;
		}
		for(no=0; no<disp; no++) {
			res[no] = data[*cur_inx + no];
		}
		res[no] = 0;
		*cur_inx = *cur_inx + disp + 1;
		tmp_inx = strlen(res) - 1;
		if(tmp_inx > 0 && res[tmp_inx] == '\n')
			res[tmp_inx--] = 0;
		return res;
	}
}


int
read_and_send_ldif_results(
    Backend	*be,
    Connection	*conn,
    Operation	*op,
    FILE	*fp,
    char	**attrs,
    int		attrsonly
)
{
	int	bsize, len;
	char	*buf, *bp;
	char	line[BUFSIZ];
	Entry	*e;
	int	err;
	char	*matched, *info;

	/* read in the result and send it along */
	buf = (char *) ch_malloc( BUFSIZ );
	buf[0] = '\0';
	bsize = BUFSIZ;
	bp = buf;
	while ( fgets( line, sizeof(line), fp ) != NULL ) {
		Debug( LDAP_DEBUG_SHELL, "shell search reading line (%s)\n",
		    line, 0, 0 );
		/* ignore lines beginning with DEBUG: */
		if ( strncasecmp( line, "DEBUG:", 6 ) == 0 ) {
			continue;
		}
		len = strlen( line );
		while ( bp + len - buf > bsize ) {
			size_t offset = bp - buf;
			bsize += BUFSIZ;
			buf = (char *) ch_realloc( buf, bsize );
			bp = &buf[offset];
		}
		strcpy( bp, line );
		bp += len;

		/* line marked the end of an entry or result */
		if ( *line == '\n' ) {
			if ( strncasecmp( buf, "RESULT", 6 ) == 0 ) {
				break;
			}

			if ( (e = str2entry( buf )) == NULL ) {
				Debug( LDAP_DEBUG_ANY, "str2entry(%s) failed\n",
				    buf, 0, 0 );
			} else {
				send_search_entry( be, conn, op, e, attrs,
				    attrsonly, NULL );
				entry_free( e );
			}

			bp = buf;
		}
	}
	(void) str2result( buf, &err, &matched, &info );

	/* otherwise, front end will send this result */
	if ( err != 0 || op->o_tag != LDAP_REQ_BIND ) {
		send_ldap_result( conn, op, err, matched, info, NULL, NULL );
	}

	free( buf );

	return( err );
}

void
print_ldif_suffixes(
    FILE        *fp,
    Backend     *be
)
{
	int     i;

	for ( i = 0; be->be_suffix[i] != NULL; i++ ) {
		fprintf( fp, "suffix: %s\n", be->be_suffix[i] );
	}
}


