/* ldif.h - ldif backend header file */
/* $OpenLDAP: pkg/ldap/servers/slapd/back-ldif/ldif.h,v 1.7.2.2 2000/06/13 17:57:37 kurt Exp $ */
/*
 * Copyright 1998-2000 The OpenLDAP Foundation, All Rights Reserved.
 * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
 */

#ifndef SLAPD_LDIF_H
#define SLAPD_LDIF_H

#include "external.h"

LDAP_BEGIN_DECL

#define BUFF_SIZE	8192
#define LDIF_DATA_SIZE	512
#define MAX_ARGS_NO	512
#define MAX_OBJCLASS_NO	4

#define EXEC		0
#define DYNAMIC		1
#define STATIC		2
#define INVALID		3

struct ldifinfo {
	char	**si_bind;	/* cmd + args to exec for bind	  */
	char	**si_unbind;	/* cmd + args to exec for unbind  */
	char	**si_search;	/* cmd + args to exec for search  */
	char	**si_compare;	/* cmd + args to exec for compare */
	char	**si_modify;	/* cmd + args to exec for modify  */
	char	**si_modrdn;	/* cmd + args to exec for modrdn  */
	char	**si_add;	/* cmd + args to exec for add	  */
	char	**si_delete;	/* cmd + args to exec for delete  */
	char	**si_abandon;	/* cmd + args to exec for abandon */
};

struct slap_backend_db;
struct slap_conn;
struct slap_op;

extern pid_t ldif_forkandexec LDAP_P((
	char **args,
	FILE **rfp,
	FILE **wfp));

extern void print_ldif_suffixes LDAP_P((
	FILE *fp,
	struct slap_backend_db *bd));

extern int read_and_send_ldif_results LDAP_P((
	struct slap_backend_db *bd,
	struct slap_conn *conn,
	struct slap_op *op,
	FILE *fp,
	char **attrs,
	int attrsonly));

extern char *ldif_getstrent(FILE *);
extern char *ldif_getdatastr(FILE *);
extern char *ldif_getdataent(char *, int *);
extern char *ldif_get_attribute(char *, char *, int *);
extern char *ldif_strip_comments(char *);
extern char *ldif_canonicalize_dn(char *);
extern char **ldif_get_environs();
extern char **ldif_get_args(char *, char *, char *);

typedef struct gris_data_obj{
       Entry *e;
       struct gris_data_obj *next;
}DataObj_gris;

typedef struct _active_ojb {
	char *dn;
	char **objectclass;
	char *base;
	int type;
	char *path;
	char **args;
	int  cachetime;
	int  timelimit;
	int  tlimit;
	int  stoptime;
	int  sizelimit;
	int  slimit;
	int lastmodified;
        int mod;
        int current;
        DataObj_gris **dataobj_gris;
	int activeobject;
	int activesearch;
	ldap_pvt_thread_mutex_t active_obj_mutex;
} ActiveObj;
LDAP_END_DECL

typedef struct _ldif_data {
	ActiveObj		***glist;
	int			total_inx;
	ldap_pvt_thread_mutex_t	glist_mutex;
	char			*ldifconfigfile;
	char			*ldifcmdadd;
	char			*ldifgiven_sfx;
	int			ldifallow_anonym;
	int			local_port;
	int			allow_stale_data;
	Entry			*stub_entry;
} LdifDT;

#endif



