/* config.c - giis backend configuration file routine */
/*
$Header: /home/globdev/CVS/globus-packages/mds/openldap/giis-ldapmodules/config.c,v 1.5 2003/03/27 03:19:11 jslee Exp $
$Log: config.c,v $
Revision 1.5  2003/03/27 03:19:11  jslee
removing duplicated lines

Revision 1.4  2003/03/22 02:50:59  jslee
adding allowstaledata yes/no

Revision 1.3  2003/02/05 09:11:10  jslee
fixing threads runout - bugzilla #644

Revision 1.2  2002/08/14 22:40:27  benc
Merge MDS2.2 into trunk.

Revision 1.1.1.1.28.1  2002/08/07 03:56:01  benc
Merged (hopefully) all of the mds_2_2_branch work onto this new branch.

Compiles and runs on Linux.

Revision 1.1.1.1.10.1  2002/05/20 22:17:16  jslee
adding syslog messages to the giis backend

Revision 1.1.1.1  2002/01/10 02:10:09  jslee
importing globus-beta-branch to the main trunk

Revision 1.2.4.5  2001/10/26 08:40:40  jslee
modifying search_fill variable

Revision 1.2.4.4  2001/10/26 02:36:40  jslee
fixing memory leaks

Revision 1.2.4.3  2001/10/23 02:47:13  jslee
debugging the initialization of giisallow_anonym value

Revision 1.2.4.2  2001/10/13 06:11:41  jslee
enabling multiple giis backends per slapd\nremoving all the global variables

Revision 1.2.4.1  2001/10/06 00:49:09  jslee
adding dn relabling function

Revision 1.2  2001/09/06 00:51:40  jslee
add policy checking & modifying timeout checking

Revision 1.2  2001/08/20 05:19:42  june
add purge & registration policy functions

Revision 1.1.1.1  2001/07/21 02:18:06  june
Import giis backend

Revision 1.1.1.1  2001/07/21 01:53:49  june
Import giis backend

Revision 1.1.1.1  2001/07/06 21:11:15  mei
1) splitting out back-giis from the openldap 2.0.8-patched into a standalone
   directory that can generate 'dynamic modules' that can be used by the
   openldap's slapd (checked against 2.0.11 already)

Revision 1.3  2001/05/24 01:05:06  june
gridmap checking & timelimit debugging

Revision 1.2  2001/04/04 21:42:07  june
debugging add operation

Revision 1.1  2001/03/24 00:08:36  june
adding giis backend

Revision 1.1  2001/03/22 03:01:25  june
add giis backend

Revision 1.3  2000/09/13 02:11:50  june
resolve changes between 2.0 & 2.0.1

Revision 1.2  2000/09/12 22:41:39  june
Add giis backend
*/



#include "portable.h"

#include <stdio.h>

#include <ac/socket.h>
#include <ac/string.h>
#include <ac/time.h>

#include "slap.h"
#include "external.h"
#include "slapd_giis.h"


int
giis_back_db_config(
    BackendDB	*be,
    const char	*fname,
    int		lineno,
    int		argc,
    char	**argv
)
{
	GiisDT *gdt = (GiisDT *) be->be_private;

	if (gdt == NULL) {
		gdt = (GiisDT *) calloc(1, sizeof(GiisDT));
		if (gdt == NULL) {
			fprintf(stderr, "giis data memory allocation error\n");
			Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
			return -1;
		}
		if( init_giisdt(gdt) == -1 )
			return -1;
		be->be_private = gdt;
	}

	if (gdt->giisgiven_sfx != NULL) free(gdt->giisgiven_sfx);
	if ( (gdt->giisgiven_sfx = (char *) strdup(be->be_suffix[0])) == NULL) {
		fprintf(stderr, "memory allocation error\n");
		Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
		return( 1 );
	}

	/* giis resource script */
	if ( strcasecmp( argv[0], "anonymousbind" ) == 0 ) {
		if ( argc < 2 ) {
			fprintf( stderr,
		"%s: line %d: missing command name in \"anonymousbinding <cmdname>\" line\n",
			    fname, lineno );
			Debug( LDAP_DEBUG_ANY, "%s: line %d: missing command name in \"anonymousbinding <cmdname>\" line\n", fname, lineno, 0 );
			return( 1 );
		}
		if (argv[1] && !strcasecmp(argv[1], "no"))
			gdt->giisallow_anonym = 0;
		else
			gdt->giisallow_anonym = 1;

	} else if ( strcasecmp( argv[0], "allowstaledata" ) == 0 ) {
		if ( argc < 2 ) {
			fprintf( stderr,
		"%s: line %d: missing command name in \"allowstaledata <cmdname>\" line\n",
			    fname, lineno );
			Debug( LDAP_DEBUG_ANY, "%s: line %d: missing command name in \"allowstaledata <cmdname>\" line\n", fname, lineno, 0 );
			return( 1 );
		}
		if (argv[1] && !strcasecmp(argv[1], "no"))
			gdt->allow_stale_data = 0;
		else
			gdt->allow_stale_data = 1;

	} else if ( strcasecmp( argv[0], "conf" ) == 0 ) {
		if ( argc < 2 ) {
			fprintf( stderr,
		"%s: line %d: missing filename in \"conf <filename>\" line\n",
			    fname, lineno );
			Debug( LDAP_DEBUG_ANY, "%s: line %d: missing filename in \"conf <filename>\" line\n", fname, lineno, 0 );
			return( 1 );
		}
		if (gdt->giisconfigfile) free(gdt->giisconfigfile);
		gdt->giisconfigfile = ch_strdup( argv[1] );
		if (!gdt->giisconfigfile) {
			fprintf(stderr, "memory allocation error\n");
			Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
			return 1;
		}
		if (readconf(gdt) == -1) {
			fprintf(stderr, "config file (%s) read error\n", gdt->giisconfigfile);
			Debug(LDAP_DEBUG_ANY, "config file (%s) read error\n", gdt->giisconfigfile, 0, 0 );
			return 1;
		}

	} else if ( strcasecmp( argv[0], "policyfile" ) == 0 ) {
		if ( argc < 2 ) {
			fprintf( stderr,
		"%s: line %d: missing filename in \"policyfile <filename>\" line\n",
			    fname, lineno );
			Debug( LDAP_DEBUG_ANY, "%s: line %d: missing filename in \"policyfile <filename>\" line\n", fname, lineno, 0 );
			return( 1 );
		}
		if (gdt->giispolicyfile) free(gdt->giispolicyfile);
		gdt->giispolicyfile = ch_strdup( argv[1] );
		if (!gdt->giispolicyfile) {
			fprintf(stderr, "memory allocation error\n");
			Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
			return 1;
		}
		if (read_policy_file(gdt) == -1) {
			fprintf(stderr, "config file (%s) read error\n", gdt->giispolicyfile);
			Debug(LDAP_DEBUG_ANY, "config file (%s) read error\n", gdt->giispolicyfile, 0, 0 );
			return 1;
		}

	/* anything else */
	} else {
		fprintf( stderr,
"%s: line %d: unknown directive \"%s\" in giis database definition (ignored)\n",
		    fname, lineno, argv[0] );
		Debug( LDAP_DEBUG_ANY, "%s: line %d: unknown directive \"%s\" in giis database definition (ignored)\n", fname, lineno, argv[0] );
	}

	return( 0 );
}

int init_giisdt(GiisDT *gdt)
{

	gdt->glist = (ActiveObj ***) calloc(1, sizeof(ActiveObj **));
	if(!gdt->glist) {
		fprintf(stderr, "memory allocation error\n");
		Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
		return -1;
	}
	*gdt->glist = NULL;
	gdt->total_inx = 0;
	ldap_pvt_thread_mutex_init( &(gdt->glist_mutex) );

	gdt->gtrailer = (TrailerObj **) calloc(1, sizeof(TrailerObj *));
	if(!gdt->gtrailer) {
		fprintf(stderr, "memory allocation error\n");
		Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
		return -1;
	}
	*gdt->gtrailer = NULL;
	gdt->ltrailer = (TrailerObj **) calloc(1, sizeof(TrailerObj *));
	if(!gdt->ltrailer) {
		fprintf(stderr, "memory allocation error\n");
		Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
		return -1;
	}
	*gdt->ltrailer = NULL;
	ldap_pvt_thread_mutex_init( &(gdt->gtrailer_mutex) );

	gdt->rpolicy = (char ***) calloc(1, sizeof(char **));
	if(!gdt->rpolicy) {
		fprintf(stderr, "memory allocation error\n");
		Debug( LDAP_DEBUG_ANY, "memory allocation error\n", 0, 0, 0 );
		return -1;
	}
	*gdt->rpolicy = NULL;
	gdt->total_rpolicy = 0;
	ldap_pvt_thread_mutex_init( &(gdt->rpolicy_mutex) );

	gdt->giisconfigfile = NULL;
	gdt->giispolicyfile = NULL;
	gdt->giiscmdadd = NULL;
	gdt->giisgiven_sfx = NULL;
	gdt->giisallow_anonym = 1;
	gdt->allow_stale_data = 1;
	gdt->local_port = get_local_port();

	return 0;
}

