/* fork.c - fork and exec a process, connecting stdin/out w/pipes */
/*
$Header: /home/globdev/CVS/globus-packages/mds/openldap/giis-ldapmodules/fork.c,v 1.2 2002/08/14 22:40:27 benc Exp $
$Log: fork.c,v $
Revision 1.2  2002/08/14 22:40:27  benc
Merge MDS2.2 into trunk.

Revision 1.1.1.1.28.1  2002/08/07 03:56:02  benc
Merged (hopefully) all of the mds_2_2_branch work onto this new branch.

Compiles and runs on Linux.

Revision 1.1.1.1.10.1  2002/07/09 05:03:53  benc
Close most file descriptors before exec'ing information providers.

Revision 1.1.1.1  2002/01/10 02:10:09  jslee
importing globus-beta-branch to the main trunk

Revision 1.1.4.1  2001/09/21 19:50:37  meisu
1) rename giis.h to slapd_giis.h
2) change the init function in init.c to match with libback_giis.la

Revision 1.1  2001/08/24 21:19:21  bletzing
New Package.

Revision 1.1.1.1  2001/07/21 02:18:06  june
Import giis backend

Revision 1.1.1.1  2001/07/21 01:53:49  june
Import giis backend

Revision 1.1.1.1  2001/07/06 21:11:15  mei
1) splitting out back-giis from the openldap 2.0.8-patched into a standalone
   directory that can generate 'dynamic modules' that can be used by the
   openldap's slapd (checked against 2.0.11 already)

Revision 1.1  2001/03/24 00:08:37  june
adding giis backend

Revision 1.1  2001/03/22 03:01:25  june
add giis backend

Revision 1.2  2000/09/23 00:08:48  june
add forkandexec

*/
/*
 * Copyright 1998-2000 The OpenLDAP Foundation, All Rights Reserved.
 * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
 */

#include "portable.h"

#include <stdio.h>

#include <ac/string.h>
#include <ac/socket.h>
#include <ac/unistd.h>

#include <ac/errno.h>

#include "slap.h"
#include "slapd_giis.h"

pid_t
forkandexec(
    char	**args,
    FILE	**rfp,
    FILE	**wfp
)
{
	int	p2c[2], c2p[2];
	pid_t	pid;
	long	maxfd;
	int	i;

	if ( pipe( p2c ) != 0 || pipe( c2p ) != 0 ) {
		Debug( LDAP_DEBUG_ANY, "pipe failed\n", 0, 0, 0 );
		return( -1 );
	}

	maxfd = sysconf(_SC_OPEN_MAX);
	
	/*
	 * what we're trying to set up looks like this:
	 *	parent *wfp -> p2c[1] | p2c[0] -> stdin child
	 *	parent *rfp <- c2p[0] | c2p[1] <- stdout child
	 */

#ifdef HAVE_THR
	switch ( (pid = fork1()) )
#else
	switch ( (pid = fork()) )
#endif
	{

	case 0:		/* child */
		/*
		 * child could deadlock here due to resources locked
		 * by our parent
		 *
		 * If so, configure --without-threads or implement forking
		 * via a surrogate parent.
		 */

		/* could possibly remove these because they'll get 
		 * closed lower down now - bc 
		close( p2c[1] );
		close( c2p[0] );
		*/

		if ( dup2( p2c[0], 0 ) == -1 || dup2( c2p[1], 1 ) == -1 ) {
			Debug( LDAP_DEBUG_ANY, "dup2 failed\n", 0, 0, 0 );
			exit( EXIT_FAILURE );
		}

		/*
		 * this loop starts at 3 because we always need to keep
		 * fds 0 and 1 open (attached to the two pipes above),
		 * and we want to keep stderr (fd 2).
		 *
		 * return code of close is deliberately not checked - 
		 * most of these are going to return
		 */

		for( i=3 ; i<maxfd ; i++ ) {
			if(close(i) == -1 && errno != EBADF )
			{
				Debug(LDAP_DEBUG_ANY, "giis forkandexec got an error closing all fds.\n", 0, 0, 0 );
			}
		}
		
		execv( args[0], args );

		Debug( LDAP_DEBUG_ANY, "execv failed\n", 0, 0, 0 );
		exit( EXIT_FAILURE );

	case -1:	/* trouble */
		Debug( LDAP_DEBUG_ANY, "fork failed\n", 0, 0, 0 );
		return( -1 );

	default:	/* parent */
		close( p2c[0] );
		close( c2p[1] );
		break;
	}

	if ( (*rfp = fdopen( c2p[0], "r" )) == NULL || (*wfp = fdopen( p2c[1],
	    "w" )) == NULL ) {
		Debug( LDAP_DEBUG_ANY, "fdopen failed\n", 0, 0, 0 );
		close( c2p[0] );
		close( p2c[1] );

		return( -1 );
	}

	return( pid );
}

char **
get_environs()
{

	return NULL;
}

