/* giis.h - giis backend header file */
/* $OpenLDAP: pkg/ldap/servers/slapd/back-giis/giis.h,v 1.7.2.2 2000/06/13 17:57:37 kurt Exp $ */
/*
 * Copyright 1998-2000 The OpenLDAP Foundation, All Rights Reserved.
 * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
 */

#ifndef SLAPD_GIIS_H
#define SLAPD_GIIS_H

#include "external.h"

LDAP_BEGIN_DECL

#define BUFF_SIZE	8192
#define GIIS_DATA_SIZE	512
#define MAX_ARGS_NO	512
#define MAX_OBJCLASS_NO	4
#define MAX_DN_LENGTH	512

#define EXEC		0
#define DYNAMIC		1
#define STATIC		2
#define INVALID		3

#define CACHE_ERROR	-1
#define CACHE_READY	0
#define CACHE_STALE	1
#define CACHE_FILL	2

#define AUTHC_ONLY	0
#define AUTHC_FIRST	1
#define ANONYM_ONLY	2

#define INIT_STEP	0
#define BINDING_STEP	1
#define SEARCHING_STEP	2

struct giisinfo {
	char	**si_bind;	/* cmd + args to exec for bind	  */
	char	**si_unbind;	/* cmd + args to exec for unbind  */
	char	**si_search;	/* cmd + args to exec for search  */
	char	**si_compare;	/* cmd + args to exec for compare */
	char	**si_modify;	/* cmd + args to exec for modify  */
	char	**si_modrdn;	/* cmd + args to exec for modrdn  */
	char	**si_add;	/* cmd + args to exec for add	  */
	char	**si_delete;	/* cmd + args to exec for delete  */
	char	**si_abandon;	/* cmd + args to exec for abandon */
};

struct slap_backend_db;
struct slap_conn;
struct slap_op;

extern pid_t forkandexec LDAP_P((
	char **args,
	FILE **rfp,
	FILE **wfp));

extern void print_giis_suffixes LDAP_P((
	FILE *fp,
	struct slap_backend_db *bd));

extern int read_and_send_giis_results LDAP_P((
	struct slap_backend_db *bd,
	struct slap_conn *conn,
	struct slap_op *op,
	FILE *fp,
	char **attrs,
	int attrsonly));

extern void  getdataent_init();
extern char *getstrent(FILE *);
extern char *getdatastr(FILE *);
extern char *getdataent(char *, int *);
extern char *get_attribute(char *, char *, int *);
extern char *strip_comments(char *);
extern char *canonicalize_dn(char *);
extern char **get_environs();
extern char **get_args(char *, char *, char *);
extern int  print_glist();
extern char *get_obj_from_entry(LDAP *, LDAPMessage *);
extern char *giis_dn_parent(char *);
extern char *giis_dn_relative(char *);
extern char *relable_dn(char *, char *);
extern char *relable_object(char *, char *);

typedef struct _data_obj {
	Entry *e;
	struct _data_obj *next;
} DataObj;

typedef struct _active_obj {
	char *dn;
	char **objectclass;
	char *base;
	int  type;
	char *path;
	char **args;
	char *hn;
	char *suffix;
	int  port;
	int  cachetime;
	int  timelimit;
	int  tlimit;
	int  stoptime;
	int  sizelimit;
	int  slimit;
	int  lastmodified;
	int  lastregistered;
	int  registerttl;
	int  state;
	int  caching_step;
	char *keepto;
	char *validto;
	int  activeobject;
	int  activesearch;
	int  regobj;
	ldap_pvt_thread_mutex_t active_obj_mutex;
	int  sbind_method;
	LDAP *ld;
	int  msgid;
	int  purged;
	int  mod;
	int  current;
	int  is_local;
	DataObj **dataobj;
	LDAPMessage *res;
} ActiveObj;

extern char *gen_reg_obj(ActiveObj *);

typedef struct _trailer_obj {
	char *dn;
	char *cdn;
	char *objectclass;
	char *objectname;
	char *lastupdate;
	char *data;
	struct _trailer_obj *next;
} TrailerObj;

typedef struct _giis_data {
	ActiveObj		***glist;
	int			total_inx;
	ldap_pvt_thread_mutex_t	glist_mutex;
	TrailerObj		**gtrailer; 
	TrailerObj		**ltrailer;
	ldap_pvt_thread_mutex_t	gtrailer_mutex;
	char			***rpolicy;
	int			total_rpolicy;
	ldap_pvt_thread_mutex_t	rpolicy_mutex;
	char			*giisconfigfile;
	char			*giispolicyfile;
	char			*giiscmdadd;
	char			*giisgiven_sfx;
	int			giisallow_anonym;
	int			always_send_cache;
	int			local_port;
	int			allow_stale_data;
} GiisDT;

LDAP_END_DECL

#endif


/*

*/
