
#include <stdlib.h>

#include <stdio.h>
#include <time.h>

#include <string.h>

#define MY_MAX_GENTIME_LEN 16

/* this function leaves off the suffix "Z" because some buggy
 * strftime() functions seem to misinterpret it as an escape */
static int
static_generalized_time (char * buffer,
			 int    max_len,
			 time_t current_utc_sec,
			 int    offset_seconds)
{
  struct tm * component_time;
  time_t utc_seconds;

  utc_seconds = current_utc_sec + offset_seconds;

  component_time = gmtime (&utc_seconds);

  return strftime (buffer, max_len, "%Y%m%d%H%M%S", component_time);
}

static void
static_print_usage (char * prog_name)
{
  fprintf (stderr, "usage: %s [-help] [-offset <seconds>]...\n", prog_name);
}


int main (int argc, char ** argv)
{
  time_t current_utc_sec;
  char buffer[MY_MAX_GENTIME_LEN];
  long int * offsets;
  int offset_cnt;
  int i;

  if (argc > 1) {
    offset_cnt = (argc - 1)/2;
    offsets = (long int *) malloc (sizeof(long int)*offset_cnt);

    for (i=1; i<argc; i++) {
      if ( strcmp (argv[i], "-help") == 0 ) {
	static_print_usage (argv[0]);
	return 0;
      }
      else if ( strcmp (argv[i], "-offset") == 0 ) {
	if ( i < (argc - 1) ) {
	  long int offset;
	  int cnt;

	  cnt = sscanf (argv[i+1], "%ld", &offset);

	  if (cnt == 1) {
	    offsets[(i-1)/2] = offset;
	  }
	  else {
	    fprintf (stderr, "non-integer argument \"%s\" to -offset\n",
		     argv[i+1]);
	    static_print_usage (argv[0]);
	    return 1;
	  }
	}
	else {
	  fprintf (stderr, "-offset requires argument\n");
	  static_print_usage (argv[0]);
	  return 1;
	}
      }
      else {
	fprintf (stderr, "unrecognized option \"%s\"\n", argv[i]);
	static_print_usage (argv[0]);
	return 1;
      }

      i++;
    }
  }
  else {
    offset_cnt = 1;
    offsets = (long int *) malloc (sizeof(long int)*offset_cnt);
    offsets[0] = 0;
  }


  current_utc_sec = time (NULL);

  for (i=0; i<offset_cnt; i++) {
    static_generalized_time (buffer, 
			     MY_MAX_GENTIME_LEN,
			     current_utc_sec,
			     offsets[i]);

    if (i>0) {
      fprintf (stdout, " %sZ", buffer);
    }
    else {
      fprintf (stdout, "%sZ", buffer);
    }
  }

  fprintf (stdout, "\n");

  return 0;
}
