/*
 * Copyright (c) 2002 Hewlett-Packard Company.  All Rights Reserved.
 *
 * Permission to use, copy, and modify this software and its documentation
 * is hereby granted under the following terms and conditions.  The above
 * copyright notice, this permission notice, and the following disclaimer
 * must appear in all copies of the software, derivative works or modified
 * versions.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND HEWLETT-PACKARD CO. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL HEWLETT-PACKARD
 * COMPANY BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <stdio.h>
#include <sys/pstat.h>

/* N.B. This program will compile in a non-AnsiC environment */

#define	SHOW_MB(_bytes)	((_bytes) / (unsigned long)(1024L * 1024L))

/*
 * Usage: meminfo_hpux_helper
 *
 * Acquire information about the VM subsystem for this machine.
 * Return results as shell variable assignments suitable for "eval".
 * Variables assigned are:
 *
 *	_mem_size_mbytes	Size of physical memory in MB.
 *	_mem_free_mbytes	Amount of available physical memory in MB.
 *	_vm_size_mbytes		Size of virtual memory in MB.
 *	_vm_free_mbytes		Amount of available virtual memory in MB.
 *
 * Always exits with 0 (success).
 */
int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct pst_static pststat;
	struct pst_dynamic pstdyn;
	struct pst_vminfo pstvm;
	unsigned long pgsz = (unsigned long)8192, val;

	/*
	 * pst_static
	 *
	 * "physical_memory" - returns the maximum number of pages
	 *    of physical memory currently installed in this machine.
	 * "page_size" - returns the number of bytes contained in one
	 *    virtual memory page on this machine.
	 */
	if (pstat_getstatic(&pststat, sizeof(pststat), 1, 0) != -1) {
		pgsz = (unsigned long)pststat.page_size;
		val = (unsigned long)pststat.physical_memory;
		printf("_mem_size_mbytes=\"%ld\"\n", SHOW_MB(val * pgsz));
	}

	/*
	 * pst_dynamic
	 *
	 * "psd_free" - physical pages that the system considers to
	 *    be unused and available to processes.
	 */
	if (pstat_getdynamic(&pstdyn, sizeof(pstdyn), 1, 0) != -1) {
		val = (unsigned long)pstdyn.psd_free;
		printf("_mem_free_mbytes=\"%ld\"\n", SHOW_MB(val * pgsz));
	}

	/*
	 * pst_vminfo
	 *
	 * "psv_swapspc_max" - max pages of on-disk backing store
	 * "psv_swapspc_cnt" - pages of on-disk backing store
	 */
	if (pstat_getvminfo(&pstvm, sizeof(pstvm), 1, 0) != -1) {
		val = (unsigned long)pstvm.psv_swapspc_max;
		printf("_vm_size_mbytes=\"%ld\"\n", SHOW_MB(val * pgsz));
		val = (unsigned long)pstvm.psv_swapspc_cnt;
		printf("_vm_free_mbytes=\"%ld\"\n", SHOW_MB(val * pgsz));
	}

	return 0;
}
